# # Code used for cases 1 and 2 in the head move red line algorithm
# # returns moves1,moves2,addedNode1,addedNode2
# def RL_Case1(N1, N2, x_1, isom_N1_N2, isom_N2_N1, randomNodes=False):
#     """
#     An implementation of Algorithm 7. Finds a sequence of head moves that makes it possible to add the highest tree node x_1 to the up-closed isomrophism.

#     :param N1: a phylogenetic network.
#     :param N2: a phylogenetic network.
#     :param x_1: a tree node in N1.
#     :param isom_N1_N2: a dictionary, containing a partial (up-closed) isomorphism map from N1 to N2. The inverse of isom_N2_N1.
#     :param isom_N2_N1: a dictionary, containing a partial (up-closed) isomorphism map from N2 to N1. The inverse of isom_N1_N2.
#     :param randomNodes: a boolean value, determining whether the random version of this algorithm is used.
#     :return: a list of head moves in N1, a list of head moves in N2, a node of N1, a node of N2. After applying the moves to the networks, the nodes can be added to the up-closed isomorphism.
#     """
#     p_1 = Parent(N1, x_1)
#     p_2 = isom_N1_N2[p_1]
#     x_2 = Child(N2, p_2, exclude=isom_N2_N1.keys(), randomNodes=randomNodes)
#     if N2.out_degree(x_2) == 2:
#         #        print("Case tree")
#         return [], [], x_1, x_2
#     elif N2.in_degree(x_2) == 2:
#         #        print("Case retic")
#         c_2 = FindTreeNode(N2, excludedSet=isom_N2_N1.keys(), randomNodes=randomNodes)
#         t_2 = Parent(N2, c_2)
#         b_2 = Child(N2, c_2, exclude=[x_2], randomNodes=randomNodes)
#         #        print(x_1,p_1,p_2,x_2,c_2,t_2,b_2)
#         # TODO: Not in the latex algo, just a minor improvement:
#         # If the other child $c_2$ of $p_2$ is a retic, then we can add it.
#         if p_2 == t_2:
#             return [], [], x_1, c_2
#         if check_movable(N2, (p_2, x_2), x_2):
#             q_2 = Parent(N2, x_2, exclude=[p_2])
#             if x_2 == t_2:
#                 #                print("(p,x) movable and x=t")
#                 return [], [((q_2, x_2), x_2, (c_2, b_2))], x_1, c_2
#             else:
#                 #                print("(p,x) movable and x!=t")
#                 return [], [((p_2, x_2), x_2, (t_2, c_2)), ((t_2, x_2), x_2, (c_2, b_2))], x_1, c_2
#         else:
#             d_2 = Child(N2, x_2)
#             z_2 = Parent(N2, x_2, exclude=[p_2])
#             # Find a leaf with parent not equal to d_2
#             l_2 = FindLeaf(N2, excludedParents=[d_2], randomNodes=randomNodes)
#             w_2 = Parent(N2, l_2)
#             #            print("other")
#             #            print(d_2,z_2,w_2,l_2)
#             if l_2 == b_2:  # after the first move, b_2 is not a child of c_2 anymore, so we fix this by taking the other child of c_2 as b_2
#                 b_2 = Child(N2, c_2, exclude=[l_2, x_2], randomNodes=randomNodes)
#             if d_2 == t_2:  # i.e., if, after the first move, x_2 is the parent of c_2
#                 return [], [((z_2, d_2), d_2, (w_2, l_2)), ((z_2, x_2), x_2, (c_2, b_2))], x_1, c_2
#             else:
#                 return [], [((z_2, d_2), d_2, (w_2, l_2)), ((p_2, x_2), x_2, (t_2, c_2)),
#                             ((t_2, x_2), x_2, (c_2, b_2))], x_1, c_2
#     else:
#         #        print("Case leaf")
#         c_2 = FindTreeNode(N2, excludedSet=isom_N2_N1.keys(), randomNodes=randomNodes)
#         t_2 = Parent(N2, c_2)

#         if p_2 == t_2:
#             #            print("no move")
#             return [], [], x_1, c_2

#             # Find a reticulation arc (s_2,r_2) that can be moved to (p_2,x_2)
#         # No randomness required, because this arc will end up at its original position again.
#         s_2 = None
#         r_2 = None
#         for node in N2.nodes():
#             if N2.in_degree(node) == 2:
#                 for parent in N2.predecessors(node):
#                     if parent != p_2 and check_movable(N2, (parent, node), node):
#                         s_2 = parent
#                         r_2 = node
#             if s_2 != None:
#                 break
#         q_2 = Parent(N2, r_2, exclude=[s_2])
#         w_2 = Child(N2, r_2)
#         #        print(p_2,x_2,t_2,c_2,s_2,q_2,r_2,w_2)
#         if r_2 == p_2:
#             #            print("r=p")
#             if s_2 == t_2:
#                 return [], [((q_2, p_2), p_2, (t_2, c_2))], x_1, c_2
#             elif q_2 == t_2:
#                 return [], [((s_2, p_2), p_2, (t_2, c_2))], x_1, c_2
#             else:
#                 return [], [((s_2, p_2), p_2, (t_2, c_2)), ((t_2, p_2), p_2, (q_2, x_2)),
#                             ((q_2, p_2), p_2, (s_2, c_2))], x_1, c_2
#         if r_2 == t_2:
#             #            print("r=t")
#             if p_2 == q_2:
#                 return [], [((s_2, r_2), r_2, (p_2, x_2))], x_1, c_2
#             else:
#                 return [], [((s_2, r_2), r_2, (p_2, x_2)), ((p_2, r_2), r_2, (q_2, c_2)),
#                             ((q_2, r_2), r_2, (s_2, x_2))], x_1, c_2
#         if s_2 != t_2:
#             #            print("s!=t")
#             return [], [((s_2, r_2), r_2, (p_2, x_2)), ((p_2, r_2), r_2, (t_2, c_2)), ((t_2, r_2), r_2, (s_2, x_2)),
#                         ((s_2, r_2), r_2, (q_2, w_2))], x_1, c_2
#         else:
#             #            print("other")
#             return [], [((s_2, r_2), r_2, (p_2, x_2)), ((p_2, r_2), r_2, (s_2, c_2)),
#                         ((s_2, r_2), r_2, (q_2, w_2))], x_1, c_2

#         # Code used for case 3 in the head move red line algorithm


# # returns moves1,moves2,addedNode1,addedNode2
# def RL_Case3(N1, N2, x_1, isom_N1_N2, isom_N2_N1, randomNodes=False):
#     """
#     An implementation of Algorithm 8. Finds a sequence of head moves that makes it possible to add the highest reticulation x_1 to the up-closed isomrophism.

#     :param N1: a phylogenetic network.
#     :param N2: a phylogenetic network.
#     :param x_1: a highest reticulation node in N1.
#     :param isom_N1_N2: a dictionary, containing a partial (up-closed) isomorphism map from N1 to N2. The inverse of isom_N2_N1.
#     :param isom_N2_N1: a dictionary, containing a partial (up-closed) isomorphism map from N2 to N1. The inverse of isom_N1_N2.
#     :param randomNodes: a boolean value, determining whether the random version of this algorithm is used.
#     :return: a list of head moves in N1, a list of head moves in N2, a node of N1, a node of N2. After applying the moves to the networks, the nodes can be added to the up-closed isomorphism.
#     """
#     p_1 = Parent(N1, x_1)
#     q_1 = Parent(N1, x_1, exclude=[p_1])
#     p_2 = isom_N1_N2[p_1]
#     cp_2 = Child(N2, p_2, exclude=isom_N2_N1.keys(), randomNodes=randomNodes)
#     q_2 = isom_N1_N2[q_1]
#     cq_2 = Child(N2, q_2, exclude=isom_N2_N1.keys(), randomNodes=randomNodes)

#     # at least one tree node, so give up
#     if N2.out_degree(cp_2) == 2 or N2.out_degree(cq_2) == 2:
#         # The proof does not provide a sequence when at least of the nodes cp_2 or cq_2 is a tree node
#         # TODO: If one of (p_2,cp_2) or (q_2,cq_2) is movable, we can still do something quite similar to what follows in the last case
#         return [], [], None, None
#     # Case 3ai
#     if cp_2 == cq_2:
#         #        print("no move")
#         return [], [], x_1, cp_2
#     # Case 3av
#     elif N2.out_degree(cp_2) == 0 and N2.out_degree(cq_2) == 0:
#         #        print("both leaves")
#         # Find a head-movable arc (s_2,r_2)
#         # As each reticulation node has a movable arc, we pick a random reticulation, and then a random movable arc incident to this node
#         r_2 = FindRetic(N2, excludedSet=isom_N2_N1.keys(), randomNodes=randomNodes)
#         s_2 = None
#         for parent in N2.predecessors(r_2):
#             if check_movable(N2, (parent, r_2), r_2):
#                 if not randomNodes:
#                     s_2 = parent
#                     break
#                 elif s_2 == None or random.getrandbits(1):
#                     s_2 = parent
#         if s_2 == p_2:
#             #            print("s=p")
#             return [], [((s_2, r_2), r_2, (q_2, cq_2)), ((q_2, r_2), r_2, (p_2, cp_2))], x_1, r_2
#         else:
#             #            print("s!=p")
#             return [], [((s_2, r_2), r_2, (p_2, cp_2)), ((p_2, r_2), r_2, (q_2, cq_2))], x_1, r_2
#     # Cases 3a(ii, iii, iv)
#     else:
#         #        print("at least one non-leaf")
#         if nx.has_path(N2, cq_2, cp_2) or N2.out_degree(cp_2) == 0:
#             # Swap p and q
#             q_2, p_2 = p_2, q_2
#             cp_2, cq_2 = cq_2, cp_2
#         #        print(p_2 , q_2, cp_2, cq_2)
#         if check_movable(N2, (p_2, cp_2), cp_2):
#             #            print("movable cp_2")
#             return [], [((p_2, cp_2), cp_2, (q_2, cq_2))], x_1, cp_2
#         else:
#             #            print("immovable cp_2")
#             z = Child(N2, cp_2)
#             t = Parent(N2, cp_2, exclude=[p_2])
#             if t == q_2:
#                 return [], [], x_1, cp_2
#             return [], [((cp_2, z), z, (q_2, cq_2)), ((t, cp_2), cp_2, (z, cq_2))], x_1, z


# # Permutes the leaves of a network so that it becomes leaf isomorphic, provided the networks were already (non-labeled) ismorphic
# def Permute_Leaves_Head(network1, network2, isom_1_2, isom_2_1, label_dict_1, label_dict_2):
#     """
#     An implementation of Algorithm 9. Determines a sequence of head moves that makes two isomorphic networks labeled isomorphic.

#     :param network1: a phylogenetic network.
#     :param network2: a phylogenetic network.
#     :param isom_1_2: a dictionary, containing a partial (up-closed) isomorphism map from network1 to network2. The inverse of isom_2_1.
#     :param isom_2_1: a dictionary, containing a partial (up-closed) isomorphism map from network2 to network1. The inverse of isom_1_2.
#     :param label_dict_1: a dictionary, containing the correspondence of nodes of network1 with the leaf labels: keys are labels and values are nodes.
#     :param label_dict_2: a dictionary, containing the correspondence of nodes of network2 with the leaf labels: keys are labels and values are nodes.
#     :return: a list of head moves that turns network1 into network2.
#     """
#     #    for i,k in isom_1_2.items():
#     #        print(i,k)
#     sequence = []
#     # labeldict[label]=leaf
#     Y = list(label_dict_1.values())
#     cycles = []
#     while len(Y) > 0:
#         y1_1 = Y.pop()
#         y_2 = isom_1_2[y1_1]
#         cycle = [y1_1]
#         while network2.node[y_2].get('label') != network1.node[cycle[0]].get('label'):
#             y_new_1 = label_dict_1[network2.node[y_2].get('label')]
#             #            if len(set(network1.predecessors(cycle[-1]))&set(network1.predecessors(y_new_1)))==0:#cycle[-1] and y_new_1 have NO common parent
#             #                cycle+=[y_new_1]
#             cycle += [y_new_1]

#             y_2 = isom_1_2[y_new_1]
#             Y.remove(y_new_1)
#         if len(cycle) > 1:
#             cycles += [list(reversed(cycle))]
#     #    print("cycles",cycles)

#     t = None
#     r = None
#     for node in network1:
#         if network1.in_degree(node) == 2:
#             for parent in network1.predecessors(node):
#                 if check_movable(network1, (parent, node), node):
#                     t = parent
#                     r = node
#         if r:
#             break
#     c_last = Child(network1, r)  # In the proof: c', the child of r in N
#     s_last = Parent(network1, r, exclude=[t])

#     for cycle in cycles:
#         c = cycle

#         # shift the cycle by one of t is the parent of the last leaf in the cycle
#         if t in network1.predecessors(cycle[-1]):
#             c = [cycle[-1]] + cycle[:-1]

#         # Skip the first move if the head r of the moving arc is already above the last leaf in the cycle
#         p_last = Parent(network1, c[-1])
#         if r != p_last:
#             move = ((t, r), r, (p_last, c[-1]))
#             sequence.append((move[0], move[1], From_Edge(network1, move[0], move[1]), move[2]))
#             network1 = DoMove(network1, move[0], move[1], move[2], check_validity=False)
#         moved_arc = (t, r)

#         c_last_before = c_last

#         for i in reversed(range(len(c))):
#             p_i = Parent(network1, r, exclude=[moved_arc[0]])
#             p_im1 = Parent(network1, c[i - 1])
#             # if p_i==p_im1, then two consecutive leaves are in a cherry, and we can skip one of them.
#             if p_i == p_im1:
#                 continue
#             #                print("do nothing, swapping a cherry")
#             else:
#                 move = ((p_i, r), r, (p_im1, c[i - 1]))
#                 sequence.append((move[0], move[1], From_Edge(network1, move[0], move[1]), move[2]))
#                 network1 = DoMove(network1, move[0], move[1], move[2], check_validity=False)
#                 moved_arc = (p_i, r)
#             # If c in the `original position' (s,c) of (t,r) is permuted by the cycle, then we need to change this original position accordingly
#             if c[i] == c_last_before:
#                 c_last = c[i - 1]

#     # Put the moving arc back to its original position
#     # The proof does this for every cycle, but can also be done once at the end
#     if (s_last, c_last) in network1.edges():
#         # (t,r) might already be back at this position
#         move = ((t, r), r, (s_last, c_last))
#         sequence.append((move[0], move[1], From_Edge(network1, move[0], move[1]), move[2]))
#         network1 = DoMove(network1, move[0], move[1], move[2], check_validity=False)

#     #    print("isomorphic", Isomorphic(network1,network2))
#     #    for s in sequence:
#     #        print(s[0])
#     return sequence


# # A heuristic for finding a head move sequence between two networks.
# def Red_Line(network1, network2):
#     """
#     An implementation of Algorithm 10. Finds a sequence of head moves from network1 to network2 by building an up-closed isomorphism.

#     :param network1: a phylogenetic network.
#     :param network2: a phylogenetic network.
#     :return: A list of head moves from network1 to network2.
#     """
#     # Find the root and labels of the networks
#     root1 = Root(network1)
#     root2 = Root(network2)
#     label_dict_1 = Labels(network1)
#     label_dict_2 = Labels(network2)

#     # initialize isomorphism
#     isom_1_2 = dict()
#     isom_1_2[root1] = root2
#     isom_2_1 = dict()
#     isom_2_1[root2] = root1
#     isom_size = 1

#     # Check if the roots are of the same type
#     if network1.out_degree(root1) != network2.out_degree(root2):
#         return False

#     # Keep track of the size of the isomorphism and the size it is at the end of the red line algorithm
#     goal_size = len(network1) - len(label_dict_1)

#     # init lists of sequence of moves
#     # list of (moving_edge,moving_endpoint,from_edge,to_edge)
#     seq_from_1 = []
#     seq_from_2 = []
#     # TODO keep track of highest nodes?

#     # Do the red line algorithm
#     while (isom_size < goal_size):
#         highest_tree_node_network1, highest_retic_network1 = HighestNodesBelow(network1, isom_1_2.keys())
#         highest_tree_node_network2, highest_retic_network2 = HighestNodesBelow(network2, isom_2_1.keys())

#         # Case1
#         if highest_tree_node_network1 != None:
#             moves_network_1, moves_network_2, added_node_network1, added_node_network2 = RL_Case1(network1, network2,
#                                                                                                   highest_tree_node_network1,
#                                                                                                   isom_1_2, isom_2_1)
#         # Case2
#         elif highest_tree_node_network2 != None:
#             moves_network_2, moves_network_1, added_node_network2, added_node_network1 = RL_Case1(network2, network1,
#                                                                                                   highest_tree_node_network2,
#                                                                                                   isom_2_1, isom_1_2)
#         # Case3
#         else:
#             moves_network_1, moves_network_2, added_node_network1, added_node_network2 = RL_Case3(network1, network2,
#                                                                                                   highest_retic_network1,
#                                                                                                   isom_1_2, isom_2_1)

#         # Now perform the moves and update the isomorphism
#         isom_1_2[added_node_network1] = added_node_network2
#         isom_2_1[added_node_network2] = added_node_network1
#         for move in moves_network_1:
#             seq_from_1.append((move[0], move[1], From_Edge(network1, move[0], move[1]), move[2]))
#             network1 = DoMove(network1, move[0], move[1], move[2], check_validity=False)
#         for move in moves_network_2:
#             seq_from_2.append((move[0], move[1], From_Edge(network2, move[0], move[1]), move[2]))
#             network2 = DoMove(network2, move[0], move[1], move[2], check_validity=False)
#         isom_size += 1
#     # TESTING FOR CORRECTNESS WHILE RUNNING
#     #        if not nx.is_isomorphic(network1.subgraph(isom_1_2.keys()),network2.subgraph(isom_2_1.keys())):
#     #            print("not unlabeled isom")
#     #            print(seq_from_1)
#     #            print(seq_from_2)
#     #            print(network1.subgraph(isom_1_2.keys()).edges())
#     #            print(network2.subgraph(isom_2_1.keys()).edges())

#     # TODO Debugging, remove after for speed
#     #    if not nx.is_isomorphic(network1,network2):
#     #        print("not unlabeled isom")
#     #        print(network1.edges())
#     #        print(network2.edges())
#     #    else:
#     #        print("unlabeled isomorphic :)")

#     # Add the leaves to the isomorphism
#     for node_1 in network1.nodes():
#         if network1.out_degree(node_1) == 0:
#             parent_1 = Parent(network1, node_1)
#             parent_2 = isom_1_2[parent_1]
#             node_2 = Child(network2, parent_2, exclude=isom_2_1.keys())
#             isom_1_2[node_1] = node_2
#             isom_2_1[node_2] = node_1

#     # Permute the leaves
#     seq_permute = Permute_Leaves_Head(network1, network2, isom_1_2, isom_2_1, label_dict_1, label_dict_2)

#     # invert seq_from_2, rename to node names of network1, and append to seq_from_1
#     return seq_from_1 + seq_permute + ReplaceNodeNamesInMoveSequence(InvertMoveSequence(seq_from_2), isom_2_1)


# def Red_Line_Random(network1, network2, repeats=1):
#     """
#     Finds a sequence of head moves from network1 to network2 by randomly building an up-closed isomorphism a number of times, and only keeping the shortest sequence.

#     :param network1: a phylogenetic network.
#     :param network2: a phylogenetic network.
#     :param repeats: an integer, determining how many repeats of Red_Line_Random_Single are performed.
#     :return: A list of head moves from network1 to network2.
#     """
#     best_seq = None
#     for i in range(repeats):
#         candidate_seq = Red_Line_Random_Single(network1, network2)
#         if best_seq == None or len(best_seq) > len(candidate_seq):
#             best_seq = candidate_seq
#     return best_seq


# def Red_Line_Random_Single(network1, network2):
#     """
#     An implementation of Algorithm 11. Finds a sequence of head moves from network1 to network2 by building randomly an up-closed isomorphism.

#     :param network1: a phylogenetic network.
#     :param network2: a phylogenetic network.
#     :return: A list of head moves from network1 to network2.
#     """
#     # Find the root and labels of the networks
#     root1 = Root(network1)
#     root2 = Root(network2)
#     label_dict_1 = Labels(network1)
#     label_dict_2 = Labels(network2)

#     # initialize isomorphism
#     isom_1_2 = dict()
#     isom_1_2[root1] = root2
#     isom_2_1 = dict()
#     isom_2_1[root2] = root1
#     isom_size = 1

#     # Check if the roots are of the same type
#     if network1.out_degree(root1) != network2.out_degree(root2):
#         return False

#     # Keep track of the size of the isomorphism and the size it is at the end of the red line algorithm
#     goal_size = len(network1) - len(label_dict_1)

#     # init lists of sequence of moves
#     # list of (moving_edge,moving_endpoint,from_edge,to_edge)
#     seq_from_1 = []
#     seq_from_2 = []
#     # TODO keep track of highest nodes?

#     # Do the red line algorithm
#     while (isom_size < goal_size):
#         highest_tree_node_network1, highest_retic_network1 = HighestNodesBelow(network1, isom_1_2.keys(), allnodes=True)
#         highest_tree_node_network2, highest_retic_network2 = HighestNodesBelow(network2, isom_2_1.keys(), allnodes=True)

#         # Construct a list of all highest nodes in a tuple with the corresponding network (in random order)
#         # I.e. If u is a highest node of network one, it will appear in the list as (u,1)
#         highest_nodes_network1 = [(u, 1) for u in highest_tree_node_network1 + highest_retic_network1]
#         highest_nodes_network2 = [(u, 2) for u in highest_tree_node_network2 + highest_retic_network2]
#         candidate_highest_nodes = highest_nodes_network1 + highest_nodes_network2
#         random.shuffle(candidate_highest_nodes)

#         # As some cases do not give an addition to the isom, we continue trying lowest nodes until we find one that does.
#         for highest_node, network_number in candidate_highest_nodes:
#             u = highest_node
#             # Case1
#             if network_number == 1 and network1.in_degree(u) == 1:
#                 #                print("Case 1")
#                 moves_network_1, moves_network_2, added_node_network1, added_node_network2 = RL_Case1(network1,
#                                                                                                       network2, u,
#                                                                                                       isom_1_2,
#                                                                                                       isom_2_1,
#                                                                                                       randomNodes=True)
#                 break
#             # Case2
#             elif network_number == 2 and network2.in_degree(u) == 1:
#                 #                print("Case 2")
#                 moves_network_2, moves_network_1, added_node_network2, added_node_network1 = RL_Case1(network2,
#                                                                                                       network1, u,
#                                                                                                       isom_2_1,
#                                                                                                       isom_1_2,
#                                                                                                       randomNodes=True)
#                 break
#             # Case3
#             elif network_number == 1 and network1.in_degree(u) == 2:
#                 #                print("Case 3")
#                 moves_network_1, moves_network_2, added_node_network1, added_node_network2 = RL_Case3(network1,
#                                                                                                       network2, u,
#                                                                                                       isom_1_2,
#                                                                                                       isom_2_1,
#                                                                                                       randomNodes=True)
#                 if added_node_network2:
#                     break
#             # Case3'
#             elif network_number == 2 and network2.in_degree(u) == 2:
#                 #                print("Case 3'")
#                 moves_network_2, moves_network_1, added_node_network2, added_node_network1 = RL_Case3(network2,
#                                                                                                       network1, u,
#                                                                                                       isom_2_1,
#                                                                                                       isom_1_2,
#                                                                                                       randomNodes=True)
#                 if added_node_network2:
#                     break

#         # Now perform the moves and update the isomorphism
#         isom_1_2[added_node_network1] = added_node_network2
#         isom_2_1[added_node_network2] = added_node_network1
#         for move in moves_network_1:
#             seq_from_1.append((move[0], move[1], From_Edge(network1, move[0], move[1]), move[2]))
#             network1 = DoMove(network1, move[0], move[1], move[2], check_validity=False)
#         for move in moves_network_2:
#             seq_from_2.append((move[0], move[1], From_Edge(network2, move[0], move[1]), move[2]))
#             network2 = DoMove(network2, move[0], move[1], move[2], check_validity=False)
#         isom_size += 1
#     # TESTING FOR CORRECTNESS WHILE RUNNING
#     #        if not nx.is_isomorphic(network1.subgraph(isom_1_2.keys()),network2.subgraph(isom_2_1.keys())):
#     #            print("not unlabeled isom")
#     #            print(seq_from_1)
#     #            print(seq_from_2)
#     #            print(network1.subgraph(isom_1_2.keys()).edges())
#     #            print(network2.subgraph(isom_2_1.keys()).edges())
#     #            print(network1.edges())
#     #            print(network2.edges())

#     # TODO Debugging, remove after for speed
#     #    if not nx.is_isomorphic(network1,network2):
#     #        print("not unlabeled isom")
#     #        print(network1.edges())
#     #        print(network2.edges())
#     #    else:
#     #        print("unlabeled isomorphic :)")

#     # Add the leaves to the isomorphism
#     for node_1 in network1.nodes():
#         if network1.out_degree(node_1) == 0:
#             parent_1 = Parent(network1, node_1)
#             parent_2 = isom_1_2[parent_1]
#             node_2 = Child(network2, parent_2, exclude=isom_2_1.keys())
#             isom_1_2[node_1] = node_2
#             isom_2_1[node_2] = node_1

#     # Permute the leaves
#     seq_permute = Permute_Leaves_Head(network1, network2, isom_1_2, isom_2_1, label_dict_1, label_dict_2)

#     # invert seq_from_2, rename to node names of network1, and append to seq_from_1
#     return seq_from_1 + seq_permute + ReplaceNodeNamesInMoveSequence(InvertMoveSequence(seq_from_2), isom_2_1)



import math

import networkx as nx

from phylox.base import find_unused_node
from phylox.cherrypicking import is_second_in_reducible_pair


def count_reducible_pairs(network):
    """
    finds the number of reducible pairs in the network
    split up by number of cherries and number of reticulated cherries.

    :param network: a phylogenetic network.
    :return: a dictionary with the number of reducible pairs in the network
        keys are "cherries" and "reticulate_cherries"

    :example:
    >>> from phylox import DiNetwork
    >>> from phylox.networkproperties.properties import count_reducible_pairs
    >>> network = DiNetwork(
    ...     edges=[(-1,0),(0,1),(1,2),(1,3),(2,3),(2,4),(3,5),(0,6),(6,7),(6,8)],
    ... )
    >>> counts = count_reducible_pairs(network)
    >>> counts["cherries"] == 1 and counts["reticulate_cherries"] == 1
    True
    """
    cherries = []
    reticulate_cherries = []
    for node in network.nodes:
        if network.out_degree(node) == 0:
            pair = is_second_in_reducible_pair(network, node)
            if pair != False:
                for parent in network.predecessors(pair[0]):
                    if network.out_degree(parent) == 1:
                        reticulate_cherries += [pair]
                        break
                else:
                    cherries += [pair]
    return {
        "cherries": len(cherries) / 2,
        "reticulate_cherries": len(reticulate_cherries),
    }


def blob_properties(network):
    """finds a list of all blobs of the network and their properties.
    Each blob is a pair (blob_size, blob_level) where blob_size is the
    number of nodes in the blob and blob_level is the number of
    reticulations in the blob.

    :param network: a phylogenetic network.
    :return: a list of pairs (blob_size, blob_level)

    :example:
    >>> from phylox import DiNetwork
    >>> from phylox.networkproperties.properties import blob_properties
    >>> network = DiNetwork(
    ...     edges=[(1,2),(2,3),(2,4),(3,4),(3,5),(4,6),(6,7),(6,8),(7,8),(7,9),(8,10)],
    ... )
    >>> blob_properties(network)
    [(3, 1), (3, 1)]

    >>> network = DiNetwork(
    ...     edges=[(0,1),(1,2),(1,3),(2,4),(3,5),(2,5),(3,4),(4,6),(5,7)],
    ... )
    >>> blob_properties(network)
    [(5, 2)]
    """
    blob_properties = []
    # For each biconnected component
    for bicomponent in nx.biconnected_components(network.to_undirected()):
        # A bicomponent is a blob if it consists of at least 2 nodes
        if len(bicomponent) > 2:
            #            blob = network.subgraph(bicomponent)
            # count reticulations in blob
            retics = 0
            for node in bicomponent:
                if network.in_degree(node) == 2:
                    retics += 1
            blob_size = len(bicomponent)
            blob_level = retics
            blob_properties += [(blob_size, blob_level)]
    return blob_properties


def level(network):
    """
    returns the level of the network

    :param network: a phylogenetic network.
    :return: the level of the network

    :example:
    >>> from phylox import DiNetwork
    >>> from phylox.networkproperties.properties import level
    >>> network = DiNetwork(
    ...     edges=[(1,2),(2,3),(2,4),(3,4),(3,5),(4,6),(6,7),(6,8),(7,8),(7,9),(8,10)],
    ... )
    >>> level(network)
    1

    >>> network = DiNetwork(
    ...     edges=[(0,1),(1,2),(1,3),(2,4),(3,5),(2,5),(3,4),(4,6),(5,7)],
    ... )
    >>> level(network)
    2
    """

    blobs = blob_properties(network)
    return max([blob[1] for blob in blobs])


def b2_balance(network, connect_roots=False):
    """returns the B_2 balance of the network

    :param network: a phylogenetic network.
    :param connect_roots: if True, connects all roots to a new root.
    :return: the B_2 balance of the network

    :example:
    >>> from phylox import DiNetwork
    >>> from phylox.networkproperties.properties import b2_balance
    >>> network = DiNetwork(
    ...     edges=[(0, 1), (1, 2), (1, 3), (2, 4), (2, 5), (3, 6), (3, 7)],
    ... )
    >>> b2_balance(network) == 2
    True
    """
    balance = 0

    # Initiate probabilities
    roots = network.roots
    if len(roots) == 0:
        return 0
    if len(roots) > 1:
        if connect_roots:
            new_root = find_unused_node(network)
            for root in roots:
                network.add_edge(new_root, root)
        else:
            raise ValueError("Network has more than one root")
        root = new_root
    else:
        root = list(roots)[0]

    probabilities = dict()
    probabilities[root] = 1
    highest_nodes = []
    for node in network.successors(root):
        if network.in_degree(node) == 1:
            highest_nodes += [node]

    # Calculate the probabilities of reaching each node with a uniform random directed walk
    while highest_nodes:
        current_node = highest_nodes.pop()
        # Calculate the probability of current_node
        prob = 0
        for parent in network.predecessors(current_node):
            prob += (1 / float(network.out_degree(parent))) * probabilities[parent]
        probabilities[current_node] = prob

        # Update the set of highest nodes
        for child in network.successors(current_node):
            for parent in network.predecessors(child):
                if parent not in probabilities:
                    break
            else:
                highest_nodes += [child]

        #
        if network.out_degree(current_node) == 0:
            balance -= prob * math.log(prob, 2)

    return balance



# # These functions are implementations of the algorithms of Remie Janssen's PhD thesis

# from __future__ import print_function
# import networkx as nx
# from RearrDistance_Tools import *
# import ast
# import os
# import sys
# import re
# import time

# ##PARAMETERS
# filename = None
# edges = False
# tailMoves = True
# headMoves = True
# bfSearch = False
# simple=False
# time_limit=False
# repeats=1
# runtime=False
# deep_dive=False
# randomNodes=False
# ##


# ###############################2. I/O############################
# option_help = False
# i = 1
# while i < len(sys.argv):
#     arg= sys.argv[i]
#     if arg == "-f" or arg == "--filename":
#         i+=1
#         filename = str(sys.argv[i])
#     if arg == "-tl" or arg == "--timelimit":
#         i+=1
#         time_limit = float(sys.argv[i])
#     if arg == "-e" or arg == "--edges":
#         edges = True
#     if arg == "-t" or arg == "--tail":
#         headMoves = False
#     if arg == "-h" or arg == "--head":
#         tailMoves = False
#     if arg == "-bfs" or arg == "--bfsearch":
#         bfSearch = True
#     if arg == "-dd" or arg == "--deepdive":
#         deep_dive = True
#     if arg == "-s" or arg == "--simple":
#         simple = True
#     if arg == "-rt" or arg == "--runtime":
#         runtime = True
#     if arg == "-r" or arg == "--random":
#         randomNodes = True
#     if arg == "-rep" or arg == "--repeats":
#         i+=1
#         repeats = int(sys.argv[i])
#     if arg == "-help" or arg == "--help":
#         option_help = True
#     i += 1

# if len(sys.argv)==1 or option_help or (deep_dive and bfSearch):
#     print("Mandatory arguments:\n -f or --filename followed by the filename of the file containing two networks \n\nOptional arguments:\n -e or --edges if the input file contains a list of edges in the form [(x1,y1),...,(xn,yn)] with xi and yi integers or strings in the form \"string\". If this option is not selected, the input is assumed to consist of two newick strings.\n -t or --tail for only using tail moves, instead of tail and head moves.\n -h or --head for only using head moves, instead of tail and head moves.\n -r or --random to pick random nodes whenever arbitrary nodes are required in the heuristics.\n -rt or --runtime to also record the runtime of the algorithm and the time to read the trees from file.\n -bfs or --bfsearch for using a breadth first search to find the an optimal sequence.\n -tl or --timelimit followed by a number of seconds to set a timelimit for the bfs. If no answer is found before the time runs out, the algorithm just returns a lower bound on the distance.\n\nThe output is given as a list of moves in the format:\n  moving_edge, moving_endpoint, to_edge")
#     sys.exit()


# ####################################################
# ####################################################
# ####################################################
# #############                          #############
# #############           MAIN           #############
# #############                          #############
# ####################################################
# ####################################################
# ####################################################


# if runtime:
#     time_start_reading = time.time()

# test = open(filename, "r")
# line1 = test.read()
# line1 = line1.split("\n")
# test.close()
# if edges:
#     N = nx.DiGraph()
#     M = nx.DiGraph()
#     N.add_edges_from(ast.literal_eval(line1[0]))
#     M.add_edges_from(ast.literal_eval(line1[1]))
#     rootN=Root(N)
#     if N.out_degree(rootN)==2:
#        N.add_edges_from([('rho',rootN)])
#     rootM=Root(M)
#     if M.out_degree(rootM)==2:
#        M.add_edges_from([('rho',rootM)])
#     N = NetworkLeafToLabel(N)
#     M = NetworkLeafToLabel(M)
#     label_attribute=None
# else:
#     N = Newick_To_Network(line1[0])
#     M = Newick_To_Network(line1[1])
#     if not simple:
#         print("The networks as list of edges, with node names as used in the computed sequence of moves")
#         print("Network 1:")
#         for e in N.edges():
#             label = N.node[e[1]].get('label')
#             if label:
#                 print(str(e[0])+" "+str(e[1])+" = leaf: "+str(label))
#             else:
#                 print(str(e[0])+" "+str(e[1]))
#         print("Network 2:")
#         for e in M.edges():
#             label = M.node[e[1]].get('label')
#             if label:
#                 print(str(e[0])+" "+str(e[1])+" = leaf: "+str(label))
#             else:
#                 print(str(e[0])+" "+str(e[1]))


# if runtime:
#     time_start_searching = time.time()


# if bfSearch:
#     if not simple:
#         print("Computing a shortest sequence using breadth first search.")
#         #Note, the code uses a DFS with incremented depth bound
#         #Otherwise, the queue gets too large for memory
#     sequence = Depth_First(N,M,tail_moves=tailMoves,head_moves=headMoves,max_time=time_limit,show_bounds=(not simple))
# elif deep_dive:
#     if not simple:
#         print("Computing a shortest sequence by passing the depth first search tree once, choosing the best branch (according to the GL algorithm) each time.")
#         #Note, the code uses a DFS with incremented depth bound
#         #Otherwise, the queue gets too large for memory
#     sequence = Deep_Dive_Scored(N,M,head_moves=headMoves)


# else:
#     if not tailMoves:
#         if not simple:
#             print("Computing a sequence using the `red-line' heuristic.")
#         if randomNodes:
#             if not simple:
#                 print("Picking random nodes whenever arbitrary nodes are required.")
#             sequence = Red_Line_Random(N,M,repeats=repeats)
#         else:
#             sequence = Red_Line(N,M)
#     else:
#         if not simple:
#             print("Computing a sequence using the `green-line' heuristic.")
#         if randomNodes:
#             if not simple:
#                 print("Picking random nodes whenever arbitrary nodes are required.")
#             sequence = Green_Line_Random(N,M,head_moves=headMoves,repeats=repeats)
#         else:
#             sequence = Green_Line(N,M,head_moves=headMoves)

# if runtime:
#     time_stop_searching = time.time()
#     if simple:
#         print(";"+str(time_start_searching-time_start_reading)+";"+str(time_stop_searching-time_start_searching),end='')
#     else:
#         print("Reading the networks cost "+str(time_start_searching-time_start_reading)+" seconds.")
#         print("Finding the sequence cost "+str(time_stop_searching-time_start_searching)+" seconds.")


# if not simple:
#     print("Sequence:")
# if sequence==False:
#     if not simple:
#         print("There is no sequence between the networks.")
#     else:
#         print(";None;None",end='')
#     sys.exit()
# if type(sequence)==int:
#     if simple:
#         print(";>="+str(sequence)+";?",end='')
#     else:
#         print("No network was found within the time limit.")
#         print("The distance between the networks is at least: "+str(sequence))
#     sys.exit()
# if edges:
#     sequence = ReplaceNodeNamesByOriginal(N,sequence)

# #Print the output


# if simple:
#     print(";"+str(len(sequence))+";"+str(sequence),end='')
# else:
#     for move in sequence:
#         if len(move)==4:
#             print(str(move[0])+" "+str(move[1])+" "+str(move[3]))
#         if len(move)==3:
#             print(str(move[0])+" "+str(move[1])+" "+str(move[2]))



import unittest

from phylox import DiNetwork
from phylox.constants import LABEL_ATTR
from phylox.isomorphism import is_isomorphic
from phylox.parser import dinetwork_to_extended_newick, extended_newick_to_dinetwork


class TestExtendedNewickToDiNetwork(unittest.TestCase):
    def test_small_tree(self):
        newick = "(a,b,(c,d));"
        network = extended_newick_to_dinetwork(newick)
        network2 = DiNetwork(
            edges=[(1, 2), (1, 3), (1, 4), (4, 5), (4, 6)],
            labels=[(2, "a"), (3, "b"), (5, "c"), (6, "d")],
        )
        self.assertTrue(is_isomorphic(network, network2))

    def test_small_tree_with_lengths(self):
        newick = "(a:1.0,b:1.1,(c:1.2,d:1.3):1.4);"
        network = extended_newick_to_dinetwork(newick)
        network2 = DiNetwork(
            edges=[(1, 2), (1, 3), (1, 4), (4, 5), (4, 6)],
            labels=[(2, "a"), (3, "b"), (5, "c"), (6, "d")],
        )
        self.assertTrue(is_isomorphic(network, network2))
        node_a = network.label_to_node_dict["a"]
        parent_a = network.parent(node_a)
        self.assertEqual(network[parent_a][node_a]["length"], 1.0)

    def test_small_tree_with_more_attrs(self):
        newick = "(a:1.0:2.0:3.0,b:1.1,(c:1.2,d:1.3):1.4);"
        network = extended_newick_to_dinetwork(newick)
        network2 = DiNetwork(
            edges=[(1, 2), (1, 3), (1, 4), (4, 5), (4, 6)],
            labels=[(2, "a"), (3, "b"), (5, "c"), (6, "d")],
        )
        self.assertTrue(is_isomorphic(network, network2))
        node_a = network.label_to_node_dict["a"]
        parent_a = network.parent(node_a)
        self.assertEqual(network[parent_a][node_a]["length"], 1.0)
        self.assertEqual(network[parent_a][node_a]["bootstrap"], 2.0)
        self.assertEqual(network[parent_a][node_a]["probability"], 3.0)

    def test_network(self):
        newick = "(a,(b)#R1,(#H1,c));"
        network = extended_newick_to_dinetwork(newick)
        network2 = DiNetwork(
            edges=[(1, 2), (1, 3), (3, 7), (1, 4), (4, 3), (4, 6)],
            labels=[(2, "a"), (7, "b"), (6, "c")],
        )
        self.assertTrue(is_isomorphic(network, network2))

    def test_network_with_lengths(self):
        newick = "(a:1.0,(b:1.1)#R1:1.2,(#H1:1.3,c:1.4):1.5);"
        network = extended_newick_to_dinetwork(newick)
        network2 = DiNetwork(
            edges=[(1, 2), (1, 3), (3, 7), (1, 4), (4, 3), (4, 6)],
            labels=[(2, "a"), (7, "b"), (6, "c")],
        )
        self.assertTrue(is_isomorphic(network, network2))
        node_a = network.label_to_node_dict["a"]
        parent_a = network.parent(node_a)
        self.assertEqual(network[parent_a][node_a]["length"], 1.0)


class TestNetworkToNewick(unittest.TestCase):
    def test_multirooted(self):
        network = DiNetwork(
            edges=[(0, 1), (2, 3), (3, 4), (3, 5)],
            labels=[(1, "a"), (4, "b"), (5, "c")],
        )
        with self.assertRaises(ValueError):
            dinetwork_to_extended_newick(network)

    def test_small_tree(self):
        network = DiNetwork(
            edges=[(1, 2), (1, 3)],
            labels=[(2, "a"), (3, "b")],
        )
        newick = dinetwork_to_extended_newick(network)
        self.assertTrue(
            newick
            in [
                "(a,b);",
                "(b,a);",
            ]
        )

    def test_small_tree_lengths(self):
        network = DiNetwork(
            edges=[
                (1, 2, {"length": 1.0}),
                (1, 3, {"length": 2.0}),
            ],
            labels=[(2, "a"), (3, "b")],
        )
        newick = dinetwork_to_extended_newick(network)
        self.assertTrue(
            newick
            in [
                "(a:1.0,b:2.0);",
                "(b:2.0,a:1.0);",
            ]
        )

    def test_small_tree_more_attrs(self):
        network = DiNetwork(
            edges=[
                (1, 2, {"length": 1.0, "bootstrap": 2.0, "probability": 3.0}),
                (1, 3, {"length": 2.0, "bootstrap": 3.0, "probability": 4.0}),
            ],
            labels=[(2, "a"), (3, "b")],
        )
        newick = dinetwork_to_extended_newick(network)
        self.assertTrue(
            newick
            in [
                "(a:1.0:2.0:3.0,b:2.0:3.0:4.0);",
                "(b:2.0:3.0:4.0,a:1.0:2.0:3.0);",
            ]
        )

    def test_network(self):
        network = DiNetwork(
            edges=[(1, 2), (1, 3), (2, 4), (3, 4), (2, 5), (3, 6), (4, 7)],
            labels=[(5, "a"), (6, "b"), (7, "c")],
        )
        newick = dinetwork_to_extended_newick(network)
        self.assertTrue(
            newick
            in [
                "((a,(c)#R0),(#H0,b));",
                "(((c)#R0,a),(#H0,b));",
                "((#H0,b),(a,(c)#R0));",
                "((#H0,b),((c)#R0,a));",
                "((a,(c)#R0),(b,#H0));",
                "(((c)#R0,a),(b,#H0));",
                "((b,#H0),(a,(c)#R0));",
                "((b,#H0),((c)#R0,a));",
                "((b,(c)#R0),(#H0,a));",
                "(((c)#R0,b),(#H0,a));",
                "((#H0,a),(b,(c)#R0));",
                "((#H0,a),((c)#R0,b));",
                "((b,(c)#R0),(a,#H0));",
                "(((c)#R0,b),(a,#H0));",
                "((a,#H0),(b,(c)#R0));",
                "((a,#H0),((c)#R0,b));",
            ]
        )

    def test_network_probability(self):
        network = DiNetwork(
            edges=[(1, 2), (1, 3), (2, 4), (3, 4), (2, 5), (3, 6), (4, 7)],
            labels=[(5, "a"), (6, "b"), (7, "c")],
        )
        network[2][4]["probability"] = 0.2
        network[3][4]["probability"] = 0.8
        newick = dinetwork_to_extended_newick(network)
        self.assertTrue(
            newick
            in [
                "((b:::,(c:::)#R0:::0.8):::,(#H0:::0.2,a:::):::);",
                "(((c:::)#R0:::0.8,b:::):::,(#H0:::0.2,a:::):::);",
                "((#H0:::0.2,a:::):::,(b:::,(c:::)#R0:::0.8):::);",
                "((#H0:::0.2,a:::):::,((c:::)#R0:::0.8,b:::):::);",
                "((b:::,(c:::)#R0:::0.8):::,(a:::,#H0:::0.2):::);",
                "(((c:::)#R0:::0.8,b:::):::,(a:::,#H0:::0.2):::);",
                "((a:::,#H0:::0.2):::,(b:::,(c:::)#R0:::0.8):::);",
                "((a:::,#H0:::0.2):::,((c:::)#R0:::0.8,b:::):::);",
            ]
        )

    def test_network_all_attrs(self):
        network = DiNetwork(
            edges=[
                (
                    1,
                    2,
                    {
                        "length": 1.0,
                        "bootstrap": 2.0,
                    },
                ),
                (1, 3, {"length": 2.0, "bootstrap": 3.0, "probability": 0.2}),
                (2, 3, {"length": 3.0, "bootstrap": 4.0, "probability": 0.8}),
                (
                    2,
                    4,
                    {
                        "length": 4.0,
                        "bootstrap": 5.0,
                    },
                ),
                (
                    3,
                    5,
                    {
                        "length": 5.0,
                        "bootstrap": 6.0,
                    },
                ),
            ],
            labels=[(4, "a"), (5, "b")],
        )
        newick = dinetwork_to_extended_newick(network)
        self.assertTrue("(b:5.0:6.0:)#R0:3.0:4.0:0.8" in newick)
        self.assertTrue("a:4.0:5.0:" in newick)
        self.assertTrue("#H0:2.0:3.0:0.2" in newick)


class TestNetworkToNewickAndBack(unittest.TestCase):
    def test_larger_network(self):
        network = DiNetwork(
            edges=[
                (1, 2),
                (1, 3),
                (2, 4),
                (3, 4),
                (2, 5),
                (3, 13),
                (4, 7),
                (7, 8),
                (7, 9),
                (8, 10),
                (9, 10),
                (8, 11),
                (9, 12),
                (10, 13),
                (13, 14),
            ],
            labels=[(5, "a"), (14, "b"), (11, "c"), (12, "d")],
        )
        newick = dinetwork_to_extended_newick(network)
        network2 = extended_newick_to_dinetwork(newick)
        self.assertTrue(is_isomorphic(network, network2))



import unittest

from phylox import DiNetwork
from phylox.constants import LABEL_ATTR
from phylox.isomorphism import is_isomorphic


class TestDiNetwork(unittest.TestCase):
    def test_init(self):
        network = DiNetwork()
        self.assertEqual(list(network.edges), [])
        self.assertEqual(list(network.nodes), [])

    def test_init_all(self):
        network = DiNetwork(
            nodes=[1, 2, 3],
            edges=[(1, 2), (2, 3)],
            labels=[(1, "a"), (2, "b"), (3, "c")],
        )
        self.assertCountEqual(list(network.nodes), [1, 2, 3])
        self.assertCountEqual(list(network.edges), [(1, 2), (2, 3)])
        self.assertEqual(network.nodes[1][LABEL_ATTR], "a")
        self.assertEqual(network.nodes[2][LABEL_ATTR], "b")
        self.assertEqual(network.nodes[3][LABEL_ATTR], "c")

    def test_leaves(self):
        network = DiNetwork(
            edges=[(1, 2), (2, 3)],
            labels=[(1, "a"), (2, "b"), (3, "c")],
        )
        self.assertEqual(network.leaves, {3})

    def test_roots(self):
        network = DiNetwork(
            edges=[(1, 2), (2, 3)],
            labels=[(1, "a"), (2, "b"), (3, "c")],
        )
        self.assertEqual(network.roots, {1})

    def test_reticulation_number(self):
        network = DiNetwork(
            edges=[(1, 2), (2, 3), (2, 4), (3, 4), (3, 5), (4, 6)],
            labels=[(5, "a"), (6, "b")],
        )
        self.assertEqual(network.reticulation_number, 1)

    def test_child(self):
        network = DiNetwork(
            edges=[(1, 2), (2, 3), (2, 4), (3, 4), (3, 5), (4, 6)],
            labels=[(5, "a"), (6, "b")],
        )
        self.assertIn(network.child(2), [3, 4])
        self.assertEqual(network.child(2, exclude=[3]), 4)
        self.assertEqual(network.child(2, exclude=[3, 4]), None)
        self.assertIn(network.child(2, randomNodes=True), [3, 4])
        self.assertEqual(network.child(2, exclude=[3], randomNodes=True), 4)
        self.assertEqual(network.child(2, exclude=[3, 4], randomNodes=True), None)

    def test_child_seed(self):
        network = DiNetwork(
            edges=[(0, i) for i in range(1, 100)],
        )
        child1 = network.child(0, randomNodes=True, seed=1)
        child2 = network.child(0, randomNodes=True, seed=1)
        child3 = network.child(0, randomNodes=True, seed=2)
        self.assertEqual(child1, child2)
        self.assertNotEqual(child1, child3)

    def test_parent(self):
        network = DiNetwork(
            edges=[(1, 2), (2, 3), (2, 4), (3, 4), (3, 5), (4, 6)],
            labels=[(5, "a"), (6, "b")],
        )
        self.assertIn(network.parent(4), [2, 3])
        self.assertEqual(network.parent(4, exclude=[2]), 3)
        self.assertEqual(network.parent(4, exclude=[2, 3]), None)
        self.assertIn(network.parent(4, randomNodes=True), [2, 3])
        self.assertEqual(network.parent(4, exclude=[2], randomNodes=True), 3)
        self.assertEqual(network.parent(4, exclude=[2, 3], randomNodes=True), None)

    def test_parent_seed(self):
        network = DiNetwork(
            edges=[(i, 0) for i in range(1, 100)],
        )
        parent1 = network.parent(0, randomNodes=True, seed=1)
        parent2 = network.parent(0, randomNodes=True, seed=1)
        parent3 = network.parent(0, randomNodes=True, seed=2)
        self.assertEqual(parent1, parent2)
        self.assertNotEqual(parent1, parent3)

    def test_is_tree_node(self):
        network = DiNetwork(
            edges=[(1, 2), (2, 3), (2, 4), (3, 4), (3, 5), (4, 6)],
            labels=[(5, "a"), (6, "b")],
        )
        self.assertFalse(network.is_tree_node(1))
        self.assertTrue(network.is_tree_node(2))
        self.assertTrue(network.is_tree_node(3))
        self.assertFalse(network.is_tree_node(4))
        self.assertFalse(network.is_tree_node(5))
        self.assertFalse(network.is_tree_node(6))

    def test_is_reticulation(self):
        network = DiNetwork(
            edges=[(1, 2), (2, 3), (2, 4), (3, 4), (3, 5), (4, 6)],
            labels=[(5, "a"), (6, "b")],
        )
        self.assertFalse(network.is_reticulation(1))
        self.assertFalse(network.is_reticulation(2))
        self.assertFalse(network.is_reticulation(3))
        self.assertTrue(network.is_reticulation(4))
        self.assertFalse(network.is_reticulation(5))
        self.assertFalse(network.is_reticulation(6))

    def test_is_leaf(self):
        network = DiNetwork(
            edges=[(1, 2), (2, 3), (2, 4), (3, 4), (3, 5), (4, 6)],
            labels=[(5, "a"), (6, "b")],
        )
        self.assertFalse(network.is_leaf(1))
        self.assertFalse(network.is_leaf(2))
        self.assertFalse(network.is_leaf(3))
        self.assertFalse(network.is_leaf(4))
        self.assertTrue(network.is_leaf(5))
        self.assertTrue(network.is_leaf(6))

    def test_is_root(self):
        network = DiNetwork(
            edges=[(1, 2), (2, 3), (2, 4), (3, 4), (3, 5), (4, 6)],
            labels=[(5, "a"), (6, "b")],
        )
        self.assertTrue(network.is_root(1))
        self.assertFalse(network.is_root(2))
        self.assertFalse(network.is_root(3))
        self.assertFalse(network.is_root(4))
        self.assertFalse(network.is_root(5))
        self.assertFalse(network.is_root(6))

    def test_from_newick(self):
        network = DiNetwork.from_newick("((a,b),c);")
        network2 = DiNetwork(
            edges=[(1, 2), (1, 3), (2, 4), (2, 5)],
            labels=[(3, "c"), (4, "b"), (5, "a")],
        )
        self.assertTrue(is_isomorphic(network, network2))

    def test_to_newick(self):
        network = DiNetwork.from_newick("(a,b);")
        newick = network.newick()
        self.assertTrue(newick in ["(a,b);", "(b,a);"])



from copy import deepcopy

from phylox.constants import LABEL_ATTR, LENGTH_ATTR


def find_unused_node(network, exclude=[]):
    """
    Find an unused node in the network.

    Parameters
    ----------
    network : networkx.DiGraph
        The network to find an unused node in.
    exclude : list
        A list of additional nodes to exclude from the search.

    Returns
    -------
    int
        The unused node.

    Examples
    --------
    >>> import networkx as nx
    >>> import phylox
    >>> network = nx.DiGraph()
    >>> network.add_edges_from([(0, 1), (1, 2), (2, 3)])
    >>> phylox.find_unused_node(network)
    -1
    >>> phylox.find_unused_node(network, exclude=[-1])
    -2
    """

    new_node = -1
    while new_node in network.nodes or new_node in exclude:
        new_node -= 1
    return new_node


def suppress_node(network, node):
    """
    Suppresses a degree-2 node node and returns true if successful.
    The new arc has length length(p,node)+length(node,c).
    Returns false if node is not a degree-2 node.

    Parameters
    ----------
    network : phylox.DiNetwork
        The network to suppress a node in.
    node : str or int
        The node to suppress.

    Returns
    -------
    bool
        True if the node was suppressed, False otherwise.

    Examples
    --------
    >>> from phylox import DiNetwork
    >>> network = DiNetwork()
    >>> network.add_edges_from([(0, 1, {'length': 1}), (1, 2, {'length': 1})])
    >>> suppress_node(network, 1)
    True
    >>> network.edges(data=True)
    OutEdgeDataView([(0, 2, {'length': 2})])
    """
    if not (network.out_degree(node) == 1 and network.in_degree(node) == 1):
        return False
    parent = network.parent(node)
    child = network.child(node)
    network.add_edges_from([(parent, child, network[parent][node])])
    if LENGTH_ATTR in network[parent][node] and LENGTH_ATTR in network[node][child]:
        network[parent][child][LENGTH_ATTR] = (
            network[parent][node][LENGTH_ATTR] + network[node][child][LENGTH_ATTR]
        )
    network.remove_node(node)
    return True


def remove_unlabeled_leaves(network, inplace=True):
    """
    Iteratively removes unlabeled leaves until none are left, then suppresses all degree-2 nodes.

    Parameters
    ----------
    network : phylox.DiNetwork
        The network to remove unlabeled leaves from.
    inplace : bool
        Whether to modify the network in place or return a copy.

    Returns
    -------
    phylox.DiNetwork
        The network with unlabeled leaves removed.

    Examples
    --------
    >>> from phylox import DiNetwork
    >>> from phylox.constants import LABEL_ATTR
    >>> network = DiNetwork(
    ...     edges=[(0, 1), (1, 2), (1, 3), (3, 4), (3, 5)],
    ...     labels=[(2, 'a'), (4, 'b')],
    ... )
    >>> network = remove_unlabeled_leaves(network)
    >>> network.edges()
    OutEdgeView([(0, 1), (1, 2), (1, 4)])
    >>> network.nodes(data=True)
    NodeDataView({0: {}, 1: {}, 2: {'label': 'a'}, 4: {'label': 'b'}})
    """
    if not inplace:
        network = deepcopy(network)

    nodes_to_check = set(deepcopy(network.nodes))
    while nodes_to_check:
        v = nodes_to_check.pop()
        if network.out_degree(v) == 0 and LABEL_ATTR not in network.nodes[v]:
            for p in network.predecessors(v):
                nodes_to_check.add(p)
            network.remove_node(v)
    list_nodes = deepcopy(network.nodes)
    for v in list_nodes:
        suppress_node(network, v)
    return network



# import CPH
# import os
# import sys
# import csv

# errorInFile = False

# #####################################################################
# #####################################################################
# ##############          Terminal arguments                ###########
# #####################################################################
# #####################################################################

# #Set default values
# option_out = False
# option_out_argument = ""
# option_file = False
# option_file_argument = ""
# option_progress = False
# option_bestTree = False
# option_repeats = False
# option_track = False
# option_repeats_argument = 1
# option_lengths = False
# option_timeLimit = False
# option_timeLimit_argument = None
# option_scoreSubnetwork = False
# option_scoreSubnetwork_argument = None
# option_reticsSubnetwork = False
# option_reticsSubnetwork_argument = None
# option_help = False

# #Read the arguments
# i = 1
# while i < len(sys.argv):
#     arg= sys.argv[i]
#     if arg == "-f" or arg == "--file":
#         option_file = True
#         i+=1
#         option_file_argument = sys.argv[i]
#         option_out_argument = option_file_argument+"OUT"
#     if arg == "-o" or arg == "--output":
#         option_out = True
#         i+=1
#         option_out_argument = sys.argv[i]
#     if arg == "-p" or arg == "--progress":
#         option_progress = True
#     if arg == "-t" or arg == "--track":
#         option_track = True
#     if arg == "-l" or arg == "--lengths":
#         option_lengths = True
#     if arg == "-r" or arg == "--repeats":
#         option_repeats = True
#         i+=1
#         option_repeats_argument = int(sys.argv[i])
#     if arg == "-rsn" or arg == "--reticssubnetwork":
#         option_reticsSubnetwork = True
#         i+=1
#         option_reticsSubnetwork_argument = int(sys.argv[i])
#     if arg == "-ssn" or arg == "--scoresubnetwork":
#         option_scoreSubnetwork = True
#         i+=1
#         option_scoreSubnetwork_argument = float(sys.argv[i])
#     if arg == "-tl" or arg == "--timelimit":
#         option_timeLimit = True
#         i+=1
#         option_timeLimit_argument = float(sys.argv[i])
#     if arg == "-bt" or arg == "--besttree":
#         option_bestTree = True
#     if arg == "-h" or arg == "--help":
#         option_help = True
#     i += 1

# #Output the help text to the terminal if no argument is given, or if the help option is chosen.
# if len(sys.argv)==1 or option_help:
#     print("Mandatory arguments:\n -f or --file followed by the input file in simple Newick format. \n\nOptional arguments:\n -o or --output followed by the output file containing the sequence, Default: [input_filename]+OUT\n -p or --progress for extra progress reports during the running of the algorithm\n -r or --repeats followed by an integer, the number of repeated attempts to find a good sequence.\n -tl or --timelimit followed by a float number of seconds, the timelimit given to the algorithm. The code will finish a repetition it is currently executing.\n -bt or --besttree for extracting the best tree out of the network, i.e. selecting the most used edges per reticulation node.\n -t or --track For tracking all reducible pairs during the algorithm.\n -l or --lengths For picking the lowest reducible cherry every time. If there are multiple, pick one at random.\n -rsn or --reticssubnetwork followed by an integer, the reticulation number of the subnetwork, to extract a subnetwork with this number of reticulations by iteratively adding reticulations with high probability (if probabilities are given) or by adding the reticulation edge that is used most by all input trees. \n -ssn or --scoresubnetwork followed by a threshold number (float) to extract a subnetwork by using all edges with a score above this threshold. If the network has probabilities, this score is the probability of an edge, otherwise it is the fraction of the input trees that uses this edge in their embedding.")
#     sys.exit()


# #####################################################################
# #####################################################################
# ##############             Read the input                 ###########
# #####################################################################
# #####################################################################

# #Empty set of inputs
# inputs = []

# #Read each line of the input file with name set by "option_file_argument"
# f = open("./"+option_file_argument, "rt")
# reader = csv.reader(f, delimiter='~', quotechar='|')
# for row in reader:
#     inputs.append(str(row[0]))
# f.close()

# #Make the set of inputs usable for all algorithms: use the CPH class
# tree_set = CPH.Input_Set(newick_strings = inputs)


# #####################################################################
# #####################################################################
# ##############            Find the sequence               ###########
# #####################################################################
# #####################################################################

# #Run the heuristic to find a cherry-pciking sequence `seq' for the set of input trees.
# #Arguments are set as given by the terminal arguments
# seq = tree_set.CPSBound(repeats = option_repeats_argument,
#                     progress    = option_progress,
#                     track       = option_track,
#                     lengths     = option_lengths,
#                     time_limit  = option_timeLimit_argument)


# #####################################################################
# #####################################################################
# ##############                   Output                   ###########
# #####################################################################
# #####################################################################


# #Open the output file for writing
# f= open(option_out_argument,"w+")

# #Output the raw network.
# f.write("Best network in Newick format;\r\n")
# if option_lengths:
#     network = CPH.PhN(seq = tree_set.best_seq_with_lengths,
#             reduced_trees = tree_set.best_seq_with_lengths_red_trees,
#             heights       = tree_set.best_seq_with_lengths_heights)
# else:
#     network = CPH.PhN(seq = tree_set.best_seq,
#             reduced_trees = tree_set.best_red_trees)
# #Cut the network into a tree to find the newick string
# cuttree = CPH.CutTree(network    = network.nw,
#                     current_node = network.no_nodes,
#                     leaf_labels  = network.leaf_nodes)
# f.write(cuttree.Newick(probabilities = True)+"\r\n")
# f.write(cuttree.Newick(probabilities = False)+"\r\n")

# #Output the desired subnetworks.
# f.write("Good subnetworks in Newick format;\r\n")
# #Find the best tree inside the network
# if option_bestTree:
#     resNW = network.SelectSubNetworkByReticulations(type_is_probability=True, reticulations = 0)
#     resNW.Clean_Up()
#     cuttree = CPH.CutTree(network = resNW.nw, current_node = resNW.no_nodes, leaf_labels  = resNW.leaf_nodes)
#     f.write(cuttree.Newick(probabilities = True)+"\r\n")
#     f.write(cuttree.Newick(probabilities = False)+"\r\n")
# #Find the best subnetwork with the given number of reticulations: option_reticsSubnetwork_argument from option -rsn
# if option_reticsSubnetwork:
#     #probabilities on edges
#     resNW = network.SelectSubNetworkByReticulations(type_is_probability=True, reticulations = option_reticsSubnetwork_argument)
#     cuttree = CPH.CutTree(network = resNW.nw, current_node = resNW.no_nodes, leaf_labels = resNW.leaf_nodes)
#     f.write(cuttree.Newick(probabilities = True)+"\r\n")
#     f.write(cuttree.Newick(probabilities = False)+"\r\n")
#     #no probabilities on edges
#     resNW = network.SelectSubNetworkByReticulations(type_is_probability=False, reticulations = option_reticsSubnetwork_argument)
#     cuttree = CPH.CutTree(network = resNW.nw, current_node = resNW.no_nodes, leaf_labels = resNW.leaf_nodes)
#     f.write(cuttree.Newick(probabilities = True)+"\r\n")
#     f.write(cuttree.Newick(probabilities = False)+"\r\n")
# #Find the best subnetwork with the given number of reticulations: option_reticsSubnetwork_argument from option -rsn
# if option_scoreSubnetwork:
#     #probabilities on edges
#     resNW = network.SelectSubNetworkByScore(type_is_probability=True, score = option_scoreSubnetwork_argument)
#     cuttree = CPH.CutTree(network = resNW.nw, current_node = resNW.no_nodes, leaf_labels = resNW.leaf_nodes)
#     f.write(cuttree.Newick(probabilities = True)+"\r\n")
#     f.write(cuttree.Newick(probabilities = False)+"\r\n")
#     #no probabilities on edges
#     resNW = network.SelectSubNetworkByScore(type_is_probability=False, score = option_scoreSubnetwork_argument)
#     cuttree = CPH.CutTree(network = resNW.nw, current_node = resNW.no_nodes, leaf_labels = resNW.leaf_nodes)
#     f.write(cuttree.Newick(probabilities = True)+"\r\n")
#     f.write(cuttree.Newick(probabilities = False)+"\r\n")


# #Output additional information about the given arguments
# f.write("END\r\n")
# f.write("Selected by score: \r\n")
# f.write(str(option_scoreSubnetwork_argument)+"\r\n")
# f.write("Selected by reticulation number: \r\n")
# f.write(str(option_reticsSubnetwork_argument)+"\r\n\r\n")

# #Output the computation time for the heuristic
# f.write("Computation time; CPS Heuristic reps\r\n")
# f.write(str(tree_set.CPS_Compute_Time) +";" + str(tree_set.CPS_Compute_Reps)+"\r\n\r\n")

# #Output the raw sequence found by the heuristic
# last = '0'
# f.write("Best sequence\r\n")
# f.write("Pair; reduced trees; (height of reduced pair)\r\n")
# for i, pair in enumerate(seq):
#     if option_lengths:
#         f.write("("+pair[0]+","+pair[1]+"); "+str(tree_set.best_seq_with_lengths_red_trees[i])+"); "+str(tree_set.best_seq_with_lengths_heights[i])+"\r\n")
#     else:
#         f.write("("+pair[0]+","+pair[1]+"); "+str(tree_set.best_red_trees[i])+"\r\n")
#     last = pair[1]
# f.write("("+last+",-)\r\n")

# #close the output file
# f.close()



from copy import deepcopy
from enum import Enum

from phylox.base import find_unused_node, suppress_node
from phylox.constants import LABEL_ATTR, LENGTH_ATTR


class CHERRYTYPE(Enum):
    CHERRY = 1
    RETICULATEDCHERRY = 2
    NONE = 0


def find_all_reducible_pairs(network):
    """
    Finds all reducible pairs (cherries and reticulated cherries) in the
    network.

    :param network: a phylogenetic network.
    :return: a set of reducible pairs (cherries and reticulated cherries) in the network.

    :example:
    >>> from phylox import DiNetwork
    >>> from phylox.cherrypicking.base import find_all_reducible_pairs
    >>> network = DiNetwork(
    ...     edges=[(-1,0),(0,1),(1,2),(1,3),(2,3),(2,4),(3,5),(0,6),(6,7),(6,8)],
    ... )
    >>> reducible_pairs = find_all_reducible_pairs(network)
    >>> reducible_pairs == {(7,8),(8,7),(5,4)}
    True
    """
    reducible_pairs = set()
    for l in network.leaves:
        reducible_pairs = reducible_pairs.union(
            find_reducible_pairs_with_second(network, l)
        )
    return reducible_pairs


def find_reducible_pairs_with_second(N, x):
    """
    Finds a list of reducible pairs (cherries and reticulated cherries) in the
    network N with leaf x as second element of the pair.

    :param N: a phylogenetic network.
    :param x: a leaf of the network N.
    :return: a list of reducible pairs (cherries and reticulated cherries) in the network N with leaf x as second element of the pair.

    :example:
    >>> from phylox import DiNetwork
    >>> from phylox.cherrypicking.base import find_reducible_pairs_with_second
    >>> network = DiNetwork(
    ...     edges=[(-1,0), (0,1), (0,2), (1,2), (1,3), (2,4)],
    ... )
    >>> find_reducible_pairs_with_second(network, 3)
    [(4, 3)]
    """
    if not N.is_leaf(x):
        raise ValueError("x must be a leaf of N")

    parent = N.parent(x)
    if N.in_degree(parent) == 0:
        return []

    reducible_pairs = list()
    for sibling in N.successors(parent):
        if sibling == x:
            continue
        sibling_out_degree = N.out_degree(sibling)
        if sibling_out_degree == 0:
            reducible_pairs.append((sibling, x))
        if sibling_out_degree == 1:
            sibling_child = N.child(sibling)
            if N.out_degree(sibling_child) == 0:
                reducible_pairs.append((sibling_child, x))
    return reducible_pairs


def find_reducible_pairs_with_first(N, x):
    """
    Finds a list of reducible pairs (cherries and reticulated cherries) in the
    network N with leaf x as first element of the pair.

    :param N: a phylogenetic network.
    :param x: a leaf of the network N.
    :return: a list of reducible pairs (cherries and reticulated cherries) in the network N with leaf x as first element of the pair.

    :example:
    >>> from phylox import DiNetwork
    >>> from phylox.cherrypicking.base import find_reducible_pairs_with_first
    >>> network = DiNetwork(
    ...     edges=[(-1,0), (0,1), (0,2), (1,2), (1,3), (2,4)],
    ... )
    >>> find_reducible_pairs_with_first(network, 4)
    [(4, 3)]
    """
    if not N.is_leaf(x):
        raise ValueError("x must be a leaf of N")

    parent = N.parent(x)

    if N.is_tree_node(parent):
        return find_cherries_with_first(N, x)
    if N.is_reticulation(parent):
        return find_reticulated_cherry_with_first(N, x)
    else:
        return []


def find_reticulated_cherry_with_first(N, x):
    """
    Finds a list of reticulated cherries in the network N with leaf x as first
    element of the pair.

    Parameters
    ----------
    N : phylox.DiNetwork
        The network to find reticulated cherries in.
    x : str or int
        The leaf to find reticulated cherries with.

    Returns
    -------
    list
        A list of reticulated cherries in the network with leaf x as first
        element of the pair.
    """

    if not N.is_leaf(x):
        raise ValueError("x must be a leaf of N")

    parent = N.parent(x)
    if parent is None:
        return []
    if not N.is_reticulation(parent):
        return []

    reticulated_cherries = list()
    for pp in N.predecessors(parent):
        for ppc in N.successors(pp):
            if ppc == parent or not N.is_leaf(ppc):
                continue
            reticulated_cherries.append((x, ppc))
    return reticulated_cherries


def find_cherries_with_first(network, x):
    """
    Finds a set of cherries in the network N with leaf x as first element of
    the pair.

    Parameters
    ----------
    network : phylox.DiNetwork
        The network to find cherries in.
    x : str or int
        The leaf to find cherries with.

    Returns
    -------
    set
        A set of cherries in the network with leaf x as first element of the
        pair.
    """
    cherries = set()
    parent = network.parent(x)
    for sibling in network.successors(parent):
        if sibling in network.leaves and sibling != x:
            cherries.add((x, sibling))
    return cherries


def is_second_in_reducible_pair(network, x):
    for node in network.predecessors(x):
        px = node
    for cpx in network.successors(px):
        if cpx != x:
            if network.out_degree(cpx) == 0:
                return (cpx, x)
            if network.out_degree(cpx) == 1:
                for ccpx in network.successors(cpx):
                    if network.out_degree(ccpx) == 0:
                        return (ccpx, x)
    return False


def reduce_pair(network, x, y, inplace=False, nodes_by_label=False):
    """
    Reduces the reducible pair (x,y) in the network.
    Note: Cache of network properties is not updated.

    Parameters
    ----------
    network : phylox.DiNetwork
        The network to reduce the reducible pair in.
    x : str or int
        The first element of the reducible pair.
    y : str or int
        The second element of the reducible pair.
    inplace : bool
        If True, the network is modified in place.
    nodes_by_label : bool
        If True, the nodes x and y are interpreted as labels.

    Returns
    -------
    phylox.DiNetwork
        The network with the reducible pair reduced.
    CHERRYTYPE
        The type of the reducible pair.

    Raises
    ------
    ValueError
        If x or y are not in the network.

    Examples
    --------
    >>> from phylox import DiNetwork
    >>> from phylox.cherrypicking.base import reduce_pair, CHERRYTYPE
    >>> network = DiNetwork(
    ...     edges=[(-1,0), (0,1), (0,2), (1,2), (1,3), (2,4)],
    ... )
    >>> network, cherry_type = reduce_pair(network, 4, 3)
    >>> cherry_type == CHERRYTYPE.RETICULATEDCHERRY
    True
    >>> set(network.edges) == {(-1, 0), (0, 3), (0, 4)}
    True
    """

    if not inplace:
        network = deepcopy(network)
    if nodes_by_label:
        x = network.labels[x][0]
        y = network.labels[y][0]

    cherry_type = check_reducible_pair(network, x, y)
    if cherry_type == CHERRYTYPE.CHERRY:
        px = network.parent(x)
        network.remove_node(x)
        suppress_node(network, px)
    if cherry_type == CHERRYTYPE.RETICULATEDCHERRY:
        px = network.parent(x)
        py = network.parent(y)
        network.remove_edge(py, px)
        suppress_node(network, px)
        suppress_node(network, py)
    return network, cherry_type


def check_reducible_pair(network, x, y):
    """
    Checks whether the pair (x,y) is a reducible pair in the network.

    :param network: a phylogenetic network.
    :param x: a leaf of the network.
    :param y: a leaf of the network.
    :return: the type of reducible pair (x,y) in the network.

    :example:
    >>> from phylox import DiNetwork
    >>> from phylox.cherrypicking.base import check_reducible_pair, CHERRYTYPE
    >>> network = DiNetwork(
    ...     edges=[(-1,0), (0,1), (0,2), (1,2), (1,3), (2,4)],
    ... )
    >>> check_reducible_pair(network, 4, 3) == CHERRYTYPE.RETICULATEDCHERRY
    True
    """
    if not network.has_node(x):
        return CHERRYTYPE.NONE
    if not network.has_node(y):
        return CHERRYTYPE.NONE
    px = network.parent(x)
    py = network.parent(y)
    if px == py:
        return CHERRYTYPE.CHERRY
    if network.out_degree(px) == 1:
        if px in network.successors(py):
            return CHERRYTYPE.RETICULATEDCHERRY
    return CHERRYTYPE.NONE


def add_pair(network, x, y, height=[1, 1], inplace=False, nodes_by_label=False):
    """
    Adds a pair to the network, using the construction from a cherry-picking sequence
    :param x: first element of the pair
    :param y: second element of the pair
    :param height: height of the pair
    :param inplace: if true, the network is modified in place, otherwise a copy is returned
    :param nodes_by_label: if true, the nodes are indexed by their label, otherwise by their index
    :return: the network with the pair added
    """
    if not inplace:
        network = deepcopy(network)

    # if the network is empty, create a cherry (x,y)
    if len(network.leaves) == 0:

        node_x = 2 if nodes_by_label else x
        node_y = 3 if nodes_by_label else y
        root = find_unused_node(network, exclude=[node_x, node_y])
        parent = find_unused_node(network, exclude=[root, node_x, node_y])
        network.add_weighted_edges_from(
            [
                (root, parent, 0),
                (parent, node_x, height[0]),
                (parent, node_y, height[1]),
            ],
            weight=LENGTH_ATTR,
        )
        if nodes_by_label:
            network.label_to_node_dict[x] = node_x
            network.label_to_node_dict[y] = node_y
            network.nodes[node_x][LABEL_ATTR] = x
            network.nodes[node_y][LABEL_ATTR] = y
        network._clear_cached()
        return network

    node_y = network.label_to_node_dict.get(y) if nodes_by_label else y
    # if y is not in the network raise an error, as there is no way to add the pair and get a phylogenetic network
    if node_y is None or node_y not in network.leaves:
        raise ValueError("y is not in the network")
    # else, add the pair to the existing network
    # get edge data for edges around y
    parent_node_y = network.parent(node_y)
    length_incoming_y = network[parent_node_y][node_y].get(LENGTH_ATTR)
    # no_of_trees_incoming_y = network[parent_node_y][node_y].get("no_of_trees")
    edge_data = dict()
    height_goal_x = height[0]
    if length_incoming_y is not None:
        if height[1] < length_incoming_y:
            height_pair_y_real = height[1]
        else:
            height_pair_y_real = length_incoming_y
            height_goal_x += height[1] - height_pair_y_real
        edge_data[LENGTH_ATTR] = height_pair_y_real
    # if no_of_trees_incoming_y is not None:
    #     edge_data["no_of_trees"] = no_of_trees_incoming_y + len(red_trees - current_trees)

    old_edge_data = network.edges[parent_node_y, node_y]
    old_edge_data[LENGTH_ATTR] = length_incoming_y - height_pair_y_real

    # add all edges around y
    network.remove_edge(parent_node_y, node_y)
    new_parent_of_y = find_unused_node(network)
    network.add_edges_from(
        [
            (parent_node_y, new_parent_of_y, old_edge_data),
            (new_parent_of_y, node_y, edge_data),
        ]
    )

    # Now also add edges around x
    node_x = (
        network.label_to_node_dict.get(x, find_unused_node(network))
        if nodes_by_label
        else x
    )
    # x is not yet in the network, so make a cherry (x,y)
    if node_x not in network.leaves:
        network.add_edge(
            new_parent_of_y,
            node_x,
            # no_of_trees=len(red_trees),
            length=height_goal_x,
        )
        if nodes_by_label:
            network.nodes[node_x][LABEL_ATTR] = x
            network.label_to_node_dict[x] = node_x
        network._clear_cached()
        return network

    # x is already in the network, so create a reticulate cherry (x,y)
    parent_node_x = network.parent(node_x)
    length_incoming_x = network[parent_node_x][node_x][LENGTH_ATTR]
    # no_of_trees_incoming_x = network[parent_node_x][node_x]["no_of_trees"]
    # if x is below a reticulation, and the height of the new pair is above the height of this reticulation, add the new hybrid arc to the existing reticulation
    if network.in_degree(parent_node_x) > 1 and length_incoming_x <= height_goal_x:
        network.add_edge(
            new_parent_of_y,
            parent_node_x,
            # no_of_trees=len(red_trees),
            length=height_goal_x - length_incoming_x,
        )
        # network[parent_node_x][node_x]["no_of_trees"] += len(red_trees)
        network._clear_cached()
        return network

    # create a new reticulation vertex above x to attach the hybrid arc to
    height_pair_x = min(height_goal_x, length_incoming_x)
    new_parent_of_x = find_unused_node(network)
    old_edge_data = network.edges[parent_node_x, node_x]
    old_edge_data[LENGTH_ATTR] = length_incoming_x - height_pair_x

    network.remove_edge(parent_node_x, node_x)
    network.add_edges_from(
        [
            (parent_node_x, new_parent_of_x, old_edge_data),
            (
                new_parent_of_x,
                node_x,
                {LENGTH_ATTR: length_incoming_x - height_pair_x},
            ),  # "no_of_trees": no_of_trees_incoming_x + len(red_trees)
            (
                new_parent_of_y,
                new_parent_of_x,
                {LENGTH_ATTR: height_goal_x - height_pair_x},
            ),  # "no_of_trees": len(red_trees)
        ]
    )
    network._clear_cached()
    return network


# TODO: make work with cps with labels instead of node indices
def get_indices_of_reducing_pairs(sequence, network):
    """
    Checks which pairs of a sequence actually reduce a given network
    for a given cherry-picking sequence `sequence' reduces a given tree `tree'
    input:
        sequence: a list of pairs of leaves
        network: a network
    output:
        if the network is reduced by the sequence, returns the list of all indices of pairs that reduce the network
        otherwise returns False
    """
    network_copy = deepcopy(network)
    indices = []
    for i, pair in enumerate(sequence):
        network_copy, cherry_type = reduce_pair(network_copy, *pair)
        if cherry_type != CHERRYTYPE.NONE:
            indices += [i]
        if len(network_copy.edges) <= 1:
            return indices
    return False


def add_roots_to_sequence(sequence, reduced_trees_per_pair):
    """
    Modifies a cherry-picking sequence so that it represents a network with exactly one root.
    A sequence may be such that reconstructing a network from the sequence results in multiple roots
    This function adds some pairs to the sequence so that the network has a single root.
    args:
        sequence: the sequence to modify
        reduced_trees_per_pair: the sets of trees reduced by each pair in the sequence
    returns:
        the new sequence, and also the sets of trees reduced by each pair in the sequence, modified so that the new pairs are also represented (they reduce no trees)
    """
    leaves_encountered = set()
    roots = set()
    # The roots can be found by going back through the sequence and finding pairs where the second element has not been encountered in the sequence yet
    for pair in reversed(sequence):
        if pair[1] not in leaves_encountered:
            roots.add(pair[1])
        leaves_encountered.add(pair[0])
        leaves_encountered.add(pair[1])
    i = 0
    roots = list(roots)
    # Now add some pairs to make sure each second element is already part of some pair in the sequence read backwards, except for the last pair in the sequence
    for i in range(len(roots) - 1):
        sequence.append((roots[i], roots[i + 1]))
        # none of the trees are reduced by the new pairs.
        reduced_trees_per_pair.append(set())
        i += 1
    return sequence, reduced_trees_per_pair


def has_cherry(network, x, y):
    """
    Checks whether the pair (x,y) forms a cherry in the network

    Parameters
    ----------
    network : phylox.DiNetwork
        The network in which we want to check whether (x,y) is a cherry
    x : string
        The first element of the pair
    y : string
        The second element of the pair

    Returns
    -------
    bool
        True if (x,y) is a cherry in the network, False otherwise
    """
    if (not x in network.leaves) or (not y in network.leaves):
        return False
    px = network.parent(x)
    py = network.parent(y)
    return px == py


def cherry_height(network, x, y):
    """
    Returns the height of (x,y) if it is a cherry:
        i.e.: length(p,x)+length(p,y)/2
    Returns false otherwise

    Parameters
    ----------
    network : phylox.DiNetwork
        The network in which we want to check the height of cherry (x,y)
    x : string
        The first element of the pair
    y : string
        The second element of the pair

    Returns
    -------
    float
        The height of the cherry (x,y) if it is a cherry, False otherwise
    """
    print("cherry_height")
    print(network.edges(data=True))

    if (not x in network.leaves) or (not y in network.leaves):
        return False
    px = network.parent(x)
    py = network.parent(y)
    if px == py:
        height = [network[px][x][LENGTH_ATTR], network[py][y][LENGTH_ATTR]]
        return height
    if (py, px) in network.edges:
        height = [
            network[px][x][LENGTH_ATTR] + network[py][px][LENGTH_ATTR],
            network[py][y][LENGTH_ATTR],
        ]
        return height
    raise ValueError("x and y are not in the same cherry")


class CherryPickingMixin:
    @classmethod
    def from_cherry_picking_sequence(cls, sequence, heights=None, label_leaves=True):
        network = cls()
        heights = heights or [[1, 1]] * len(sequence)
        for pair, height in zip(reversed(sequence), reversed(heights)):
            add_pair(
                network, *pair, height=height, inplace=True, nodes_by_label=label_leaves
            )
        network._clear_cached()
        return network



import random

import networkx as nx
from networkx.utils.decorators import np_random_state, py_random_state

from phylox.cherrypicking.base import CherryPickingMixin
from phylox.constants import LABEL_ATTR, LENGTH_ATTR


class DiNetwork(nx.DiGraph, CherryPickingMixin):
    """
    A class for representing a directed phylogenetic network.
    Inherits from networkx.DiGraph.

    :param edges: a list of edges of the network.
    :param nodes: a list of nodes of the network.
    :param labels: a list of tuples of the form (node, label) where node is a node of the network and label is a label of the node.
    :param kwargs: additional keyword arguments.
    """

    def __init__(self, *args, **kwargs):
        edges = kwargs.get("edges", [])
        super().__init__(edges, *args, **kwargs)
        self.add_nodes_from(kwargs.get("nodes", []))
        self.label_tuples = kwargs.get("labels", [])
        for label_tuple in self.label_tuples:
            self.add_node(label_tuple[0], label=label_tuple[1])

    def _clear_cached(self):
        """
        Clears all cached properties of the network.

        :return: None
        """
        for attr in [
            "_leaves",
            "_reticulations",
            "_roots",
            "_reticulation_number, _labels, _label_to_node_dict",
        ]:
            if hasattr(self, attr):
                delattr(self, attr)

    @classmethod
    def from_newick(cls, newick):
        """
        Creates a network from a newick string.
        Not implemented.

        :param newick: a newick string.
        :return: a network.
        """
        from phylox.parser import extended_newick_to_dinetwork

        return extended_newick_to_dinetwork(newick)

    def _set_leaves(self):
        """
        Sets the set of leaves of the network as a cached property.

        :return: the set of leaves of the network.
        """
        self._leaves = set([node for node in self.nodes if self.is_leaf(node)])
        return self._leaves

    def _set_label_to_node_dict(self):
        """
        Sets the dictionary mapping labels to nodes as a cached property.

        :return: the dictionary mapping labels to nodes.
        """
        self._label_to_node_dict = {}
        for node in self.nodes:
            if LABEL_ATTR in self.nodes[node]:
                self._label_to_node_dict[self.nodes[node][LABEL_ATTR]] = node
        return self._label_to_node_dict

    @property
    def label_to_node_dict(self):
        """
        Returns the dictionary mapping labels to nodes.
        Uses a cached property if available.

        :return: the dictionary mapping labels to nodes.
        """
        if not hasattr(self, "_label_to_node_dict"):
            self._set_label_to_node_dict()
        return self._label_to_node_dict

    @property
    def leaves(self):
        """
        Returns the set of leaves of the network.
        Uses a cached property if available.

        :return: the set of leaves of the network.
        """
        if not hasattr(self, "_leaves"):
            self._set_leaves()
        return self._leaves

    def _set_reticulations(self):
        self._reticulations = set(
            [node for node in self.nodes if self.is_reticulation(node)]
        )
        return self._reticulations

    @property
    def reticulations(self):
        """
        Returns the set of reticulations of the network.
        Uses a cached property if available.

        :return: the set of reticulations of the network.
        """
        if not hasattr(self, "_retculations"):
            self._set_reticulations()
        return self._reticulations

    def _set_roots(self):
        """
        Sets the set of roots of the network as a cached property.

        :return: the set of roots of the network.
        """
        self._roots = set([node for node in self.nodes if self.is_root(node)])
        return self._roots

    @property
    def roots(self):
        """
        Returns the set of roots of the network.
        Uses a cached property if available.

        :return: the set of roots of the network.
        """
        if not hasattr(self, "_roots"):
            self._set_roots()
        return self._roots

    @property
    def reticulation_number(self):
        """
        Returns the number of reticulations of the network.
        Uses a cached property if available.

        :return: the number of reticulations of the network.
        """
        if not hasattr(self, "_reticulation_number"):
            self._reticulation_number = sum(
                [max(self.in_degree(node) - 1, 0) for node in self.nodes]
            )
        return self._reticulation_number

    def _set_labels(self):
        """
        Sets the dictionary mapping labels to lists of nodes as a cached property.

        :return: the dictionary mapping labels to lists of nodes.
        """

        self._labels = {}
        for node in self.nodes:
            if LABEL_ATTR in self.nodes[node]:
                label = self.nodes[node][LABEL_ATTR]
                if label not in self._labels:
                    self._labels[label] = []
                self._labels[self.nodes[node][LABEL_ATTR]] += [node]
        return self._labels

    @property
    def labels(self):
        """
        Returns the dictionary mapping labels to lists of nodes.
        Use this instead of label_to_node_dict if there are multiple nodes with the same label.
        Uses a cached property if available.

        :return: the dictionary mapping labels to lists of nodes.
        """
        if not hasattr(self, "_labels"):
            self._set_labels()
        return self._labels

    @py_random_state("seed")
    def child(self, node, exclude=[], randomNodes=False, seed=None):
        """
        Finds a child node of a node.

        :param node: a node of self.
        :param exclude: a set of nodes of self.
        :param randomNodes: a boolean value.
        :param seed: a seed for the random number generator.
        :return: a child of node that is not in the set of nodes exclude. If randomNodes, then this child node is selected uniformly at random from all candidates.
        """
        child = None

        for i, c in enumerate(self.successors(node)):
            if c not in exclude:
                if not randomNodes:
                    return c
                elif child is None or seed.random() < 1.0 / (i + 1):
                    child = c
        return child

    @py_random_state("seed")
    def parent(self, node, exclude=[], randomNodes=False, seed=None):
        """
        Finds a parent of a node in a network.

        :param node: a node in the network.
        :param exclude: a set of nodes of the network.
        :param randomNodes: a boolean value.
        :param seed: a seed for the random number generator.
        :return: a parent of node that is not in the set of nodes exclude. If randomNodes, then this parent is selected uniformly at random from all candidates.
        """
        parent = None
        for i, p in enumerate(self.predecessors(node)):
            if p not in exclude:
                if not randomNodes:
                    return p
                elif parent is None or seed.random() < 1.0 / (i + 1):
                    parent = p
        return parent

    def is_reticulation(self, node):
        """
        Checks whether a node is a reticulation.
        I.e., whether it has in-degree > 1 and out-degree <= 1.

        :param node: a node in the network.
        :return: a boolean value.
        """
        return self.out_degree(node) <= 1 and self.in_degree(node) > 1

    def is_leaf(self, node):
        """
        Checks whether a node is a leaf.
        I.e., whether it has out-degree = 0 and in-degree > 0.

        :param node: a node in the network.
        :return: a boolean value.
        """
        return self.out_degree(node) == 0 and self.in_degree(node) > 0

    def is_root(self, node):
        """
        Checks whether a node is a root.
        I.e., whether it has in-degree = 0 and out-degree > 0.

        :param node: a node in the network.
        :return: a boolean value.
        """
        return self.in_degree(node) == 0

    def is_tree_node(self, node):
        """
        Checks whether a node is a tree node.
        I.e., whether it has in-degree <= 1 and out-degree > 1.

        :param node: a node in the network.
        :return: a boolean value.
        """
        return self.out_degree(node) > 1 and self.in_degree(node) <= 1

    def newick(self):
        """
        Returns a newick string representing the network.

        :return: a newick string.
        """
        from phylox.parser import dinetwork_to_extended_newick

        return dinetwork_to_extended_newick(self)



import argparse
import itertools
import math
import random
import sys

import networkx as nx
import numpy as np
import scipy.stats

# from AddEdgesToTree import *
from phylox.generators.heath.heath import RestrictToLeafSet, generate_heath_network


def parse_args():
    parser = argparse.ArgumentParser(
        description="""
        Runs a speciation-extinction-HGT-hybridization model for the given time (--time) or until a certain number of extant taxa (--taxa) is reached. If all lineages go extinct before the given time is reached, another attempt is made.
        
        Each extant taxon has its own speciation, HGT, and extinction rates (1/mean_time_until_next_event). 
        
        The hybridization rate of a pair of species is a function of the weighed distance between these species (sum of all up-down distances, weighed by their probability). 
        
        There are prior speciation/extinction/HGT rate distributions: gamma distributions with a given mean and a shape parameter for speciation (--speciation_parameters) and extinction (--extinction_parameters), HGT is turned off by default. 
        
        If a HGT event happens for a given taxon, another taxon (including itself) is chosen uniformly at random to donate genetic material (uniformly distributed contribution in [0,max_hgt] where max_hgt is determined by the --hgt_inheritance parameter). After speciation or hybridization, each rate of the new lineages is set by multiplying the (weighted mean) rate of the parent lineage(s) by a gamma-distributed factor with mean 1 and a shape parameter (--update-shape-parameter), and then accepting this rate with a probability proportional to the prior distribution for this rate. This gives an ultrametric network on the extant species.
        
        Optional arguments:
            -ti or --time followed by the total length (float) for the network. 
            -ta or --taxa followed by the number of taxa at which the simulation stops. If all lineages go extinct before the given number of taxa is reached, another attempt is made.
            -ce or --count_extinct to also count the extinct taxa as part of the taxa limit.-oe or --only_extant to return the network restricted to the extant leaves
        Rates:
            -sp or --speciation_parameters followed by a mean (float) and a shape parameter (float) for the gamma distribution of the speciation rate.
            -ext or --extinction_parameters followed by a mean (float) and a shape parameter (float) for the gamma distribution of the extinction rate.
            -noext or --no_extinction to turn off extinction altogether.
            -hgt or --hgt_parameters followed by a mean (float) and a shape parameter (float) for the gamma distribution of the HGT rate.
            -upd or --update_shape_parameter followed by a shape parameter (float) for the update gamma distribution.
            -hyb or --hyb_factor followed by four floats for the piecewise linear dependence of hybridization rate on the distance:

        left bound:  l
        right bound: r
        left rate:   rl
        right rate:  rr.
        
        |
        lr+---
        |     \
        |      \
        |       \
        rr+      -----
        |
        0 +---+----+-----
        0   l    r""",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    parser.add_argument(
        "-ti",
        "--time",
        dest="time_limit",
        type=float,
        help="Total length for the network",
    )
    parser.add_argument(
        "-ta",
        "--taxa",
        dest="taxa_limit",
        type=float,
        help="Number of taxa at which the simulation stops",
    )
    parser.add_argument(
        "-ce",
        "--count_extinct",
        action="store_true",
        help="Count extinct taxa as part of the taxa limit",
    )
    parser.add_argument(
        "-sp",
        "--speciation_parameters",
        nargs=2,
        type=float,
        metavar=("mean", "shape"),
        help="Mean and shape parameter for the gamma distribution of the speciation rate",
    )
    parser.add_argument(
        "-ext",
        "--extinction_parameters",
        nargs=2,
        type=float,
        metavar=("mean", "shape"),
        help="Mean and shape parameter for the gamma distribution of the extinction rate",
    )
    parser.add_argument(
        "-noext",
        "--no_extinction",
        action="store_true",
        help="Turn off extinction altogether",
    )
    parser.add_argument(
        "-hgt",
        "--hgt_parameters",
        nargs=2,
        type=float,
        metavar=("mean", "shape"),
        help="Mean and shape parameter for the gamma distribution of the HGT rate",
    )
    parser.add_argument(
        "--hgt_inheritance",
        type=float,
        help="Maximum contribution of a donor taxon to the genome of the recipient taxon",
    )
    parser.add_argument(
        "-hyb",
        "--hybridization_factor",
        nargs=4,
        type=float,
        metavar=("l", "r", "rl", "rr"),
        help="Piecewise linear dependence of hybridization rate on the distance",
    )
    parser.add_argument(
        "-upd",
        "--update_shape_parameter",
        type=float,
        help="Shape parameter for the update gamma distribution",
    )
    parser.add_argument(
        "-s", "--simple", action="store_true", help="Enable simple output"
    )

    args = parser.parse_args()

    params = {
        "time_limit": args.time_limit,
        "taxa_limit": args.taxa_limit,
        "count_extinct": args.count_extinct,
        "speciation_rate_mean": args.speciation_parameters[0]
        if args.speciation_parameters
        else None,
        "speciation_rate_shape": args.speciation_parameters[1]
        if args.speciation_parameters
        else None,
        "extinction_rate_mean": args.extinction_parameters[0]
        if args.extinction_parameters
        else None,
        "extinction_rate_shape": args.extinction_parameters[1]
        if args.extinction_parameters
        else None,
        "ext_used": not args.no_extinction,
        "hgt_rate_mean": args.hgt_parameters[0] if args.hgt_parameters else None,
        "hgt_rate_shape": args.hgt_parameters[1] if args.hgt_parameters else None,
        "hgt_used": args.hgt_parameters is not None,
        "hybridization_left_bound": args.hybridization_factor[0]
        if args.hybridization_factor
        else None,
        "hgt_inheritance": args.hgt_inheritance,
        "hybridization_right_bound": args.hybridization_factor[1]
        if args.hybridization_factor
        else None,
        "hybridization_left_rate": args.hybridization_factor[2]
        if args.hybridization_factor
        else None,
        "hybridization_right_rate": args.hybridization_factor[3]
        if args.hybridization_factor
        else None,
        "hyb_used": args.hybridization_factor is not None,
        "update_shape": args.update_shape_parameter,
        "simple_output": args.simple,
    }
    return params


def main():
    params = parse_args()
    # Find a network
    if params["taxa_limit"]:
        leaves = []
        no_of_extinct = 0
        while (
            len(leaves) + params["count_extinct"] * no_of_extinct
            != params["taxa_limit"]
        ):
            if not params["simple_output"]:
                print("starting over")
            network, hybrid_nodes, leaves, no_of_extinct = generate_heath_network(
                **params
            )
    else:
        print(params)
        network, hybrid_nodes, leaves, no_of_extinct = generate_heath_network(**params)

    # Restrict to extant leaves if wanted
    if only_extant:
        network = RestrictToLeafSet(network, leaves)

    for e in network.edges:
        info = ""
        if e[0] in hybrid_nodes:
            info += "H" + str(hybrid_nodes[e[0]])
        else:
            info += str(e[0])
        info += "  "
        if e[1] in hybrid_nodes:
            info += "H" + str(hybrid_nodes[e[1]])
        else:
            info += str(e[1])
        print(info)  # ,network[e[0]][e[1]]['length'])


if __name__ == "__main__":
    main()



"""
A module for generating (random) phylogenetic networks.

This implements the algorithm from the paper:
Zhang, C., Ogilvie, H.A., Drummond, A.J., Stadler, T.: Bayesian inference of species networks from multilocus
sequence data. Molecular biology and evolution 35(2), 504–517 (2018)

"""

import numpy as np
from networkx.utils.decorators import np_random_state, py_random_state

from phylox import DiNetwork
from phylox.constants import LABEL_ATTR, LENGTH_ATTR


def multree_to_dinetwork(multree, hybrid_nodes):
    """
    Converts a multree to a DiNetwork.

    Parameters
    ----------
    multree : networkx.DiGraph
        The multree to convert.
        Edges must have a LENGTH_ATTR attribute.
    hybrid_nodes : dict
        A dictionary mapping hybrid nodes to their hybrid number.

    Returns
    -------
    network : DiNetwork
        The converted network.
    """

    edges = []
    for e in multree.edges(data=True):
        node_0 = f"#H{hybrid_nodes[e[0]]}" if e[0] in hybrid_nodes else e[0]
        node_1 = f"#H{hybrid_nodes[e[1]]}" if e[1] in hybrid_nodes else e[1]
        edges.append((node_0, node_1, {LENGTH_ATTR: e[2][LENGTH_ATTR]}))
    network = DiNetwork(
        edges=edges,
    )
    return network


@np_random_state("seed")
def generate_network_zods(time_limit, speciation_rate, hybridization_rate, seed=None):
    """
    Generates a network with the ZODS model.

    Parameters
    ----------
    time_limit : float
        The time limit of the network.
    speciation_rate : float
        The speciation rate of each lineage in the network.
    hybridization_rate : float
        The hybridization rate of each pair of lineages in the network.
    seed : int or None
        The seed to use for the random number generator.

    Returns
    -------
    network : DiNetwork
        The generated network.
    """

    multree = DiNetwork()
    leaves = set([0])
    current_node = 1

    extra_time = seed.exponential(1 / float(speciation_rate))
    current_time = extra_time
    current_speciation_rate = float(speciation_rate)
    current_hybridization_rate = float(0)
    rate = current_speciation_rate + current_hybridization_rate

    # First create a MUL-tree
    hybrid_nodes = dict()
    no_of_hybrids = 0

    while current_time < time_limit:
        if seed.random() < current_speciation_rate / rate:
            # Speciate
            splitting_leaf = seed.choice(list(leaves))
            multree.add_weighted_edges_from(
                [
                    (splitting_leaf, current_node, 0),
                    (splitting_leaf, current_node + 1, 0),
                ],
                weight=LENGTH_ATTR,
            )
            leaves.remove(splitting_leaf)
            leaves.add(current_node)
            leaves.add(current_node + 1)
            current_node += 2
        else:
            # Hybridize
            #  i.e.: pick two leaf nodes, merge those, and add a new leaf below this hybrid node.
            merging = seed.choice(list(leaves), size=2)
            l0 = merging[0]
            l1 = merging[1]
            pl0 = -1
            for p in multree.predecessors(l0):
                pl0 = p
            pl1 = -1
            for p in multree.predecessors(l1):
                pl1 = p
            # If pl0==pl1, the new hybridization results in parallel edges.
            if pl0 != pl1:
                no_of_hybrids += 1
                multree.add_weighted_edges_from(
                    [(l0, current_node, 0)], weight=LENGTH_ATTR
                )
                leaves.remove(l0)
                leaves.remove(l1)
                leaves.add(current_node)
                hybrid_nodes[l0] = no_of_hybrids
                hybrid_nodes[l1] = no_of_hybrids
                current_node += 1
        # Now extend all pendant edges
        for l in leaves:
            pl = -1
            for p in multree.predecessors(l):
                pl = p
            multree[pl][l][LENGTH_ATTR] += extra_time
        no_of_leaves = len(leaves)
        current_speciation_rate = float(speciation_rate * no_of_leaves)
        current_hybridization_rate = float(
            hybridization_rate * (no_of_leaves * (no_of_leaves - 1)) / 2
        )
        rate = current_speciation_rate + current_hybridization_rate
        extra_time = seed.exponential(1 / rate)
        current_time += extra_time

    # Correct for the fact that we might have gone over the time limit
    extra_time -= current_time - time_limit
    # nothing has happened yet, and there is only one node
    if len(multree) == 0:
        multree.add_weighted_edges_from([(0, 1, time_limit)], weight=LENGTH_ATTR)
        leaves = set([1])
    # each leaf has a parent node, and we can extend each parent edge to time_limit
    else:
        multree.add_weighted_edges_from([(-1, 0, 0)], weight=LENGTH_ATTR)
        for l in leaves:
            pl = -1
            for p in multree.predecessors(l):
                pl = p
            multree[pl][l][LENGTH_ATTR] += extra_time

    # Now we have a MUL-tree, but we need to make it a network.
    return multree_to_dinetwork(multree, hybrid_nodes)



"""
Exact distance algorithms for rearrangement distances.

This module contains the exact distance algorithms for rearrangement distances.
The algorithms are based on the PhD thesis of R. Janssen, 
"Rearranging Phylogenetic Networks", 2021.
"""

from phylox.rearrangement.exact_distance.base import *



# Find the original location of the moving_endpoint,
# That is, the edge from which we remove it.
def from_edge(network, moving_edge, moving_endpoint):
    """
    Finds the original location (from-edge) of the moving_endpoint of an edge if it is moved.

    :param network: a phylogenetic network.
    :param moving_edge: an edge of the network.
    :param moving_endpoint: a node of the network, which must be an endpoint of the edge.
    :return: a pair of nodes (p,c) where p and c are a parent and child of the moving_endpoint such that they are both not part of the moving_edge.

    :example:
    >>> from phylox import DiNetwork
    >>> from phylox.rearrangement.invertsequence import from_edge
    >>> network = DiNetwork(
    ...     edges=[(0,1),(1,2),(1,3)],
    ... )
    >>> from_edge(network, (1,2), 1)
    (0, 3)
    """
    other_parent = network.parent(moving_endpoint, exclude=moving_edge)
    other_child = network.child(moving_endpoint, exclude=moving_edge)
    return (other_parent, other_child)


# def InvertMoveSequence(seq):
#     """
#     Inverts a sequence of moves.

#     :param seq: a list of moves, where each moves is in the format (moving_edge,moving_endpoint,from_edge,to_edge).
#     :return: the inverse list of moves.
#     """
#     newSeq = []
#     for move in reversed(seq):
#         moving_edge, moving_endpoint, from_edge, to_edge = move
#         newSeq.append((moving_edge, moving_endpoint, to_edge, from_edge))
#     return newSeq


# def ReplaceNodeNamesInMoveSequence(seq, isom):
#     """
#     Renames the nodes in a sequence of moves using an isomorphism mapping between two networks.

#     :param seq: a list of moves, implicitly using the nodes of a network.
#     :param isom: a dictionary, containing a bijective mapping from nodes of the networks to another set.
#     :return: a list of moves where the nodes are replaced by their image under the isom mapping.
#     """
#     if type(seq) == int:
#         return isom[seq]
#     return list(map(lambda x: ReplaceNodeNamesInMoveSequence(x, isom), seq))


# def ReplaceNodeNamesByOriginal(network, seq):
#     """
#     Renames the nodes in a sequence of moves by their original names as given in the input. These are stored as a node attribute 'original'.

#     :param network: a phylogenetic network.
#     :param seq: a sequence of moves on teh network.
#     :return: a list of moves on the network using the original node names as used in the input.
#     """
#     if type(seq) == int:
#         return network.node[seq]['original']
#     if seq == 'rho':
#         return "rho"
#     return list(map(lambda x: ReplaceNodeNamesByOriginal(network, x), seq))



"""
A module for finding properties of phylogenetic networks. 
Such as the balance of the network, the level, and the number of cherries and reticulated cherries.

Originally written by Remie Janssen for the paper
Comparing the topology of phylogenetic network generators" 
by Remie Janssen and Pengy Liu.
"""

from phylox.networkproperties.properties import *



import random
from enum import Enum

from networkx.utils.decorators import np_random_state, py_random_state

from phylox.rearrangement.move import Move, apply_move
from phylox.rearrangement.movetype import MoveType


# Pick two edges uniformly at random and add an edge between these
@np_random_state("seed")
def random_vplu_move_at_bottom(network, seed=None):
    """
    Returns a VPLU move that adds an edge between the incoming edges of two leaves.

    :param network: the network to add an edge to.
    :param seed: the seed to use for the random number generator, may be an integer or a numpy.random.RandomState.
    :return: a VPLU move that adds an edge between the incoming edges of two leaves.

    :example:
    >>> from phylox import DiNetwork
    >>> from phylox.rearrangement.movetype import MoveType
    >>> from phylox.generators.trees.well_known import generate_balanced_tree
    >>> from phylox.generators.trees.add_edges import random_vplu_move_at_bottom
    >>> tree = generate_balanced_tree(8)
    >>> move = random_vplu_move_at_bottom(tree)
    >>> move.is_type(MoveType.VPLU)
    True
    """
    leaves = list(network.leaves)
    leaf_indices = seed.choice(range(len(leaves)), 2, replace=False)
    leaf1 = leaves[leaf_indices[0]]
    parent1 = network.parent(leaf1)
    leaf2 = leaves[leaf_indices[1]]
    parent2 = network.parent(leaf2)
    return Move(
        move_type=MoveType.VPLU,
        network=network,
        start_edge=(parent1, leaf1),
        end_edge=(parent2, leaf2),
    )


# Pick two edges uniformly at random and add an edge between these
@np_random_state("seed")
def random_vplu_move_uniform(network, seed=None):
    """
    Returns a VPLU move that adds an edge between two edges in the network.
    Two edges are chosen uniformly at random from the network.

    :param network: the network to add an edge to.
    :param seed: the seed to use for the random number generator, may be an integer or a numpy.random.RandomState.
    :return: a VPLU move that adds an edge between two edges in the network.

    :example:
    >>> from phylox import DiNetwork
    >>> from phylox.rearrangement.movetype import MoveType
    >>> from phylox.generators.trees.well_known import generate_balanced_tree
    >>> from phylox.generators.trees.add_edges import random_vplu_move_uniform
    >>> tree = generate_balanced_tree(8)
    >>> move = random_vplu_move_uniform(tree)
    >>> move.is_type(MoveType.VPLU)
    True
    """
    edges = list(network.edges())
    edge_indices = seed.choice(range(len(edges)), 2, replace=False)
    return Move(
        move_type=MoveType.VPLU,
        network=network,
        start_edge=edges[edge_indices[0]],
        end_edge=edges[edge_indices[1]],
    )


@py_random_state("seed")
def random_vplu_move_local(
    network, stop_prob=0.2, max_steps=None, max_tries=None, seed=None
):
    """
    Returns a VPLU move that adds an edge between two edges in the network.
    Pick one edge, move a random number of edges through the network to find a second edge.

    :param network: the network to add an edge to.
    :param stop_prob: the probability to stop the random walk.
    :param max_steps: the maximum number of steps to take in the random walk.
    :param max_tries: the maximum number of tries to find a second edge.
    :return: a VPLU move that adds an edge between two edges in the network.

    :example:
    >>> from phylox import DiNetwork
    >>> from phylox.rearrangement.movetype import MoveType
    >>> from phylox.generators.trees.well_known import generate_balanced_tree
    >>> from phylox.generators.trees.add_edges import random_vplu_move_local
    >>> tree = generate_balanced_tree(8)
    >>> move = random_vplu_move_local(tree)
    >>> move.is_type(MoveType.VPLU)
    True
    """
    try_number = 1
    while max_tries == None or try_number <= max_tries:
        # Pick a random edge
        edge1 = seed.choice(list(network.edges()))
        edge2 = None
        # Initiate the random walk, by choosing an orientation
        previous_node = seed.choice(edge1)
        current_node = edge1[0]
        if current_node == previous_node:
            current_node = edge1[1]
        # Take a number of steps
        step_number = 1
        while max_steps == None or step_number <= max_steps:
            previous_node, current_node = current_node, seed.choice(
                list(network.successors(current_node))
                + list(network.predecessors(current_node))
            )
            if seed.random() < stop_prob:
                break
            step_number += 1
        # Set the new edge
        edge2 = (previous_node, current_node)
        if edge2 not in network.edges():
            edge2 = (current_node, previous_node)
        # Add an edge if possible, otherwise repeat the search
        if edge1 != edge2:
            break
        try_number += 1
    return Move(
        move_type=MoveType.VPLU,
        network=network,
        start_edge=edge1,
        end_edge=edge2,
    )


class AddEdgeMethod(Enum):
    UNIFORM = 1
    BOTTOM = 2
    LOCAL = 3


@np_random_state("seed")
def network_from_tree(tree, reticulations, method, seed=None):
    """
    Returns a network with the given number of reticulations added to the given tree.

    :param tree: a phylogenetic network, i.e., a DAG with leaf labels stored as the node attribute LABEL_ATTR.
    :param reticulations: the number of reticulations to add to the tree.
    :param method: the method to use to add the reticulations.
    :return: a network with the given number of reticulations added to the given tree.

    :example:
    >>> from phylox import DiNetwork
    >>> from phylox.rearrangement.movetype import MoveType
    >>> from phylox.generators.trees.well_known import generate_balanced_tree
    >>> from phylox.generators.trees.add_edges import network_from_tree, AddEdgeMethod
    >>> tree = generate_balanced_tree(8)
    >>> network = network_from_tree(tree, 2, AddEdgeMethod.BOTTOM)
    >>> len(network.leaves)
    8
    >>> network.reticulation_number
    2
    """

    if method == AddEdgeMethod.BOTTOM:
        leaves = tree.leaves

    if method == AddEdgeMethod.UNIFORM:
        add_edge_method = random_vplu_move_uniform
    elif method == AddEdgeMethod.BOTTOM:
        add_edge_method = random_vplu_move_at_bottom
    elif method == AddEdgeMethod.LOCAL:
        add_edge_method = random_vplu_move_local
    else:
        raise ValueError(f"Method {method} not implemented, use one of {AddEdgeMethod}")

    network = tree.copy()
    reticulations = int(reticulations)
    while reticulations > 0:
        try:
            move = add_edge_method(network, seed=seed)
            network = apply_move(network, move)
            reticulations -= 1
        except:
            pass
    return network



# # Add the probabilities for each edge, returns true
# # probabilties of hybrid edges are stored as edge property 'prob' (i.e. fraction of TREES THAT GO THROUGH THE RETICULATION NODE that use this edge)
# # probabilities of all edges are stored in 'probability_all'
# # the fraction of ALL INPUT TREES going through the edge is stored in 'frac_of_trees'
# def ScoreEdges(network):
#     for node in network.nodes:
#         # for the hybrid edges
#         if network.in_degree(node) > 1:
#             total_trees = 0.0
#             for parent in network.predecessors(node):
#                 total_trees += network[parent][node]["no_of_trees"]
#             for parent in network.predecessors(node):
#                 network[parent][node]["prob"] = (
#                     network[parent][node]["no_of_trees"] / total_trees
#                 )
#                 network[parent][node]["probability_all"] = network[parent][node]["prob"]
#                 if network.no_embedded_trees > 0:
#                     network[parent][node]["frac_of_trees"] = network[parent][node][
#                         "no_of_trees"
#                     ] / float(network.no_embedded_trees)
#         # and also for the non-hybrid edges (their probability is one)
#         else:
#             for parent in network.predecessors(node):
#                 network[parent][node]["probability_all"] = 1
#                 if network.no_embedded_trees > 0:
#                     network[parent][node]["frac_of_trees"] = network[parent][node][
#                         "no_of_trees"
#                     ] / float(network.no_embedded_trees)
#     return True


# # returns the 'best tree' in the network
# # i.e., pick the best incoming arc for each reticulation, and use this to find a tree.
# def Best_Tree(network):
#     edges = []
#     for v in network.nodes:
#         best_p = None
#         best_value = -1
#         for p in network.predecessors(v):
#             if network[p][v]["no_of_trees"] > best_value:
#                 best_value = network[p][v]["no_of_trees"]
#                 best_p = p
#         if best_p:
#             edges.append((best_p, v, network[best_p][v]))
#     return edges


# # Returns a subnetwork by selecting the highest scoring hybrid arcs (we pick 'reticulations' of these) and extending this to a network, so that the resulting network
# # the threshold may be the fraction of trees that uses the edge, or the probability of the edge.
# def SelectSubNetworkByReticulations(network, type_is_probability=True, reticulations=0):
#     restrictedNetwork = PhN(best_tree_from_network=network)
#     score_type = "frac_of_trees"
#     if type_is_probability:
#         score_type = "probability_all"
#     for i in range(reticulations):
#         best_score = -1
#         best_edge = None
#         for e in network.edges:
#             if (not restrictedNetwork.has_edge(e[0], e[1])) and network[e[0]][e[1]][
#                 score_type
#             ] > best_score:
#                 best_edge = e
#                 best_score = network[e[0]][e[1]][score_type]
#         # If we have already selected the whole network
#         if not best_edge:
#             break
#         restrictedNetwork.add_edges_from(
#             [(best_edge[0], best_edge[1], network[best_edge[0]][best_edge[1]])]
#         )
#         up_node = best_edge[1]
#         while restrictedNetwork.in_degree(up_node) == 0:
#             best_score = -1
#             best_parent = None
#             for parent in network.predecesors(up_node):
#                 if network[parent][up_node][score_type] > best_score:
#                     best_score = network[parent][up_node][score_type]
#                     best_parent = parent
#             restrictedNetwork.add_edges_from(
#                 [(best_parent, up_node, network[best_parent][up_node])]
#             )
#             up_node = best_parent
#     restrictedNetwork.Clean_Up()
#     # Debug: check if we find a valid network
#     restrictedNetwork.IsANetwork()
#     restrictedNetwork.ScoreEdges()
#     return restrictedNetwork


# # Returns a subnetwork by selecting all hybrid arcs that have a score above a given threshold and extending this to a network
# # the threshold may be the fraction of trees that uses the edge, or the probability of the edge.
# def SelectSubNetworkByScore(network, type_is_probability=True, score=0.5):
#     restrictedNetwork = PhN(best_tree_from_network=network)
#     score_type = "frac_of_trees"
#     if type_is_probability:
#         score_type = "probability_all"
#     done = False
#     while not done:
#         best_score = -1
#         best_edge = None
#         for e in network.edges:
#             if (not restrictedNetwork.has_edge(e[0], e[1])) and network[e[0]][e[1]][
#                 score_type
#             ] > best_score:
#                 best_edge = e
#                 best_score = network[e[0]][e[1]][score_type]
#         if best_score >= score:

#             restrictedNetwork.add_edges_from(
#                 [(best_edge[0], best_edge[1], network[best_edge[0]][best_edge[1]])]
#             )
#             up_node = best_edge[1]
#             while restrictedNetwork.in_degree(up_node) == 0:
#                 best_score = -1
#                 best_parent = None
#                 for parent in network.predecesors(up_node):
#                     if network[parent][up_node][score_type] > best_score:
#                         best_score = network[parent][up_node][score_type]
#                         best_parent = parent
#                 restrictedNetwork.add_edges_from(
#                     [(best_parent, up_node, network[best_parent][up_node])]
#                 )
#                 up_node = best_parent
#         else:
#             done = True
#     restrictedNetwork.Clean_Up()
#     # Debug: check if we find a valid network
#     restrictedNetwork.IsANetwork()
#     restrictedNetwork.ScoreEdges()
#     return restrictedNetwork



"""
A module for constructing and reducing networks using cherry-picking.

This code was initially based on a collection of scripts used for several papers:
 - "On cherry-picking and network containment" by Remie Janssen and Yukihiro Murakami (2020)
 - "Combining networks using cherry picking sequences" by Remie Janssen, Mark Jones, and Yukihiro Murakami (2020)
 - "Linear Time Algorithm for Tree-Child network Containment" by Remie Janssen and Yukihiro Murakami (2020)
 - "A practical fixed-parameter algorithm for constructing tree-child networks from multiple binary trees" by Leo van Iersel, Remie Janssen, Mark Jones, Yukihiro Murakami, and Norbert Zeh (2022)
and some unpublished work that was continued by Esther Julien:
 - Constructing Phylogenetic Networks via Cherry Picking and Machine Learning" by G BErnardini, LJJ van Iersel, E Julien, and L Stougie (2023)
"""

from .base import *



"""
A module for generating (random) phylogenetic networks.

Several types of generators have been implemented, the algorithms are based on work by other authors.
These specific implementations for (heath, lgt, mcmc, trees, zods) are adapted from the ones used for the paper
"Comparing the topology of phylogenetic network generators" 
by Remie Janssen and Pengy Liu (2021).

The randomTC generator is based on the algorithm originally by Yukihiro Murakami, and improved by Remie Janssen to create test sets for the paper
"Linear Time Algorithm for Tree-Child network Containment" by Remie Janssen and Yukihiro Murakami (2020).

The beta-splitting tree generator is adapted from a script provided by a (colleague of) Pengyu Liu.
Which in turn was based on the Beta-splitting model by Aldous (1996).

The lgt generator is based on the algorithm from:
Joan Carles Pons, Celine Scornavacca, and Gabriel Cardona. Generation of level-k LGT
networks. IEEE/ACM transactions on computational biology and bioinformatics, 17(1):158–164,
2019.

The ZODS generator is based on the algorithm from:
Zhang, C., Ogilvie, H.A., Drummond, A.J., Stadler, T.: Bayesian inference of species networks from multilocus
sequence data. Molecular biology and evolution 35(2), 504–517 (2018)
"""



# These functions are implementations of the algorithms of Remie Janssen's PhD thesis

import ast
import re
import sys
import time
from collections import deque
from copy import deepcopy

import networkx as nx

from phylox.exceptions import InvalidMoveDefinitionException, InvalidMoveException
from phylox.rearrangement.movetype import MoveType


def check_valid(network, move):
    """
    Checks whether a move is valid.

    :param move: a rearrangement move (see phylox.rearrangement.movetype.Move)
    :return: void
    :exception: InvalidMoveException if the move is not valid

    :example:
    >>> from phylox import DiNetwork
    >>> from phylox.rearrangement.move import Move
    >>> from phylox.rearrangement.movability import check_valid
    >>> network = DiNetwork(
    ...     edges=[(0,1),(1,2),(1,3),(2,3),(2,4),(3,5)],
    ... )
    >>> move = Move(
    ...     move_type=MoveType.HEAD,
    ...     origin=(2, 5),
    ...     moving_edge=(1, 3),
    ...     target=(2, 4),
    ... )
    >>> check_valid(network, move)
    """
    if move.move_type == MoveType.NONE:
        return
    if move.is_type(MoveType.RSPR):
        if not network.has_edge(
            move.origin[0], move.moving_node
        ) or not network.has_edge(move.moving_node, move.origin[1]):
            # also catches wrong endpoint type:
            # e.g.: reticulation moving_node for tail move
            raise InvalidMoveException(
                "origin does not match parent and child or moving_endpoint"
            )
        if network.has_edge(move.origin[0], move.origin[1]):
            raise InvalidMoveException("removal creates parallel edges")

        if move.is_type(MoveType.TAIL):
            if nx.has_path(network, move.moving_edge[1], move.target[0]):
                raise InvalidMoveException("reattachment would create a cycle")
            if move.target[1] == move.moving_edge[1]:
                raise InvalidMoveException("reattachment creates parallel edges")
            return
        # move.is_type(MoveType.HEAD)
        if nx.has_path(network, move.target[1], move.moving_edge[0]):
            raise InvalidMoveException("reattachment would create a cycle")
        if move.target[0] == move.moving_edge[0]:
            raise InvalidMoveException("reattachment creates parallel edges")
    elif move.is_type(MoveType.VPLU):
        if move.start_node in network.nodes:
            raise InvalidMoveException("Start node must not be in the network.")
        if move.end_node in network.nodes:
            raise InvalidMoveException("End node must not be in the network.")
        if (
            nx.has_path(network, move.end_edge[1], move.start_edge[0])
            or move.start_edge == move.end_edge
        ):
            raise InvalidMoveException("end node is reachable from start node")
    elif move.is_type(MoveType.VMIN):
        parent_0 = network.parent(move.removed_edge[0], exclude=[move.removed_edge[1]])
        child_0 = network.child(move.removed_edge[0], exclude=[move.removed_edge[1]])
        parent_1 = network.parent(move.removed_edge[1], exclude=[move.removed_edge[0]])
        child_1 = network.child(move.removed_edge[1], exclude=[move.removed_edge[0]])
        if parent_0 == parent_1 and child_0 == child_1:
            raise InvalidMoveException("removal creates parallel edges")
        if not (
            check_movable(network, move.removed_edge, move.removed_edge[0])
            and check_movable(network, move.removed_edge, move.removed_edge[1])
        ):
            raise InvalidMoveException("removal creates parallel edges")
        if network.out_degree(move.removed_edge[1]) == 0:
            raise InvalidMoveException("removes a leaf")
    else:
        raise InvalidMoveException(
            "Only rSPR and vertical moves are supported currently"
        )


# Checks whether an endpoint of an edge is movable.
def check_movable(network, moving_edge, moving_endpoint):
    """
    Checks whether an endpoint of an edge is movable in a network.

    :param network: a phylogenetic network, i.e., a DAG with labeled leaves.
    :param moving_edge: an edge in the network.
    :param moving_endpoint: a node, specifically, an endpoint of the moving_edge.
    :return: True if the endpoint of the edge is movable in the network, False otherwise.

    :example:
    >>> from phylox import DiNetwork
    >>> from phylox.rearrangement.movability import check_movable
    >>> network = DiNetwork(
    ...     edges=[(0,1),(1,2),(1,3),(2,3),(2,4),(3,5)],
    ... )
    >>> check_movable(network, (1, 3), 3)
    True
    >>> check_movable(network, (1, 3), 1)
    True
    >>> check_movable(network, (3, 5), 3)
    False
    """
    if moving_endpoint == moving_edge[0]:
        # Tail move
        if network.in_degree(moving_endpoint) in (0, 2):
            # cannot move the tail if it is a reticulation or root
            return False
    elif moving_endpoint == moving_edge[1]:
        # Head move
        if network.out_degree(moving_endpoint) in (0, 2):
            # cannot move the head if it is a tree node or leaf
            return False
    else:
        # Moving endpoint is not part of the moving edge
        return False
    # Now check for triangles, by finding the other parent and child of the moving endpoint
    parent_of_moving_endpoint = network.parent(
        moving_endpoint, exclude=[moving_edge[0]]
    )
    child_of_moving_endpoint = network.child(moving_endpoint, exclude=[moving_edge[1]])
    # if there is an edge from the parent to the child, there is a triangle
    # Otherwise, it is a movable edge
    return not network.has_edge(parent_of_moving_endpoint, child_of_moving_endpoint)



from .invalidmovedefinitionexception import InvalidMoveDefinitionException
from .invalidmoveexception import InvalidMoveException
from .nosolutionexception import NoSolutionException
from .timeoutexception import TimeoutException



"""
Heuristic distance algorithms for rearrangement distances.

This module contains the heuristic distance algorithms for rearrangement distances.
The algorithms are based on thescripts used for PhD thesis of R. Janssen, 
"Rearranging Phylogenetic Networks", 2021.

The proofs originate from constructive proofs by induction for 
bounds on the distance between two networks. These proofs can be found in
the following papers:
 - Exploring the tiers of rooted phylogenetic network space using tail moves
R Janssen, M Jones, PL Erdős, L van Iersel, C Scornavacca
Bulletin of mathematical biology 80 (8), 2177-2208
 - Heading in the right direction? Using head moves to traverse phylogenetic network space
R Janssen
Journal of Graph Algorithms and Applications 25 (1), 263-310
 - Rearranging phylogenetic networks
R Janssen
PhD thesis
"""



from copy import deepcopy

import numpy as np
from networkx.utils.decorators import np_random_state
from networkx.exception import NetworkXError

from phylox.base import find_unused_node, suppress_node
from phylox.exceptions import InvalidMoveDefinitionException, InvalidMoveException
from phylox.rearrangement.invertsequence import from_edge
from phylox.rearrangement.movability import check_valid
from phylox.rearrangement.movetype import MoveType


def apply_move(network, move):
    """
    Apply a move to the network, not in place.
    returns True if successful, and False otherwise.

    :param network: a phylogenetic network (phylox.DiNetwork).
    :param move: a move (phylox.rearrangement.move.Move) to apply to the network.
    :return: a new phylogenetic network (phylox.DiNetwork) with the move applied.

    :example:
    >>> from phylox import DiNetwork
    >>> from phylox.rearrangement.move import apply_move, Move
    >>> network = DiNetwork(
    ...     edges=[(0,1),(1,2),(1,3),(2,3),(2,4),(3,5)],
    ... )
    >>> move = Move(
    ...     move_type=MoveType.HEAD,
    ...     origin=(2,5),
    ...     moving_edge=(1,3),
    ...     target=(2,4),
    ... )
    >>> new_network = apply_move(network, move)
    >>> edges = set(new_network.edges())
    >>> edges == {(0, 1), (1, 2), (1, 3), (2, 3), (3, 4), (2, 5)}
    True
    """
    check_valid(network, move)
    new_network = deepcopy(network)

    if move.move_type in [MoveType.TAIL, MoveType.HEAD]:
        if move.moving_node in move.target:
            # move does not impact the network
            return network
        new_network.remove_edges_from(
            [
                (move.origin[0], move.moving_node),
                (move.moving_node, move.origin[1]),
                move.target,
            ]
        )
        new_network.add_edges_from(
            [
                (move.target[0], move.moving_node),
                (move.moving_node, move.target[1]),
                move.origin,
            ]
        )
        return new_network
    elif move.move_type in [MoveType.VPLU]:
        new_network.remove_edges_from([move.start_edge, move.end_edge])
        new_network.add_edges_from(
            [
                (move.start_edge[0], move.start_node),
                (move.start_node, move.start_edge[1]),
                (move.end_edge[0], move.end_node),
                (move.end_node, move.end_edge[1]),
                (move.start_node, move.end_node),
            ]
        )
        if hasattr(new_network, "_reticulation_number"):
            new_network._reticulation_number += 1
        return new_network
    elif move.move_type in [MoveType.VMIN]:
        new_network.remove_edge(*move.removed_edge)
        suppress_node(new_network, move.removed_edge[0])
        suppress_node(new_network, move.removed_edge[1])
        if hasattr(new_network, "_reticulation_number"):
            new_network._reticulation_number -= 1
        return new_network
    elif move.move_type in [MoveType.NONE]:
        return network
    raise InvalidMoveException("only tail or head moves are currently valid.")


def apply_move_sequence(network, seq_moves):
    """
    Apply a sequence of moves to the network, not in place.
    returns the resulting network.

    :param network: a phylogenetic network (phylox.DiNetwork).
    :param seq_moves: a sequence of moves (list of phylox.rearrangement.move.Move) to apply to the network.
    :return: a new phylogenetic network (phylox.DiNetwork) with the moves applied.

    :example:
    >>> from phylox import DiNetwork
    >>> from phylox.rearrangement.move import apply_move_sequence, Move
    >>> network = DiNetwork(
    ...     edges=[(0,1),(1,2),(1,3),(2,4),(2,5),(3,6),(3,7)],
    ... )
    >>> seq_moves = [
    ...     Move(
    ...         move_type=MoveType.TAIL,
    ...         moving_edge=(3,6),
    ...         target=(0,1),
    ...         origin=(1,7),
    ...     ),
    ...     Move(
    ...         move_type=MoveType.VPLU,
    ...         start_edge=(1,2),
    ...         end_edge=(1,7),
    ...         network=network,
    ...     ),
    ... ]
    >>> new_network = apply_move_sequence(network, seq_moves)
    >>> edges = set(new_network.edges())
    >>> edges == {(0, 3), (3, 1), (1, -1), (-1, 2), (1, -2), (-2, 7), (2, 4), (2, 5), (3, 6), (-1, -2)}
    True
    """

    for move in seq_moves:
        network = apply_move(network, move)
    return network


def _all_valid_tail_or_head_moves(network, move_type=MoveType.TAIL):
    if move_type not in [MoveType.TAIL, MoveType.HEAD]:
        raise InvalidMoveException("only tail or head moves are valid options.")
    if move_type == MoveType.TAIL:
        target_endpoint_index = 0
    else:
        target_endpoint_index = 1

    for moving_edge in network.edges():
        for target_edge in network.edges():
            moving_endpoint_index = target_endpoint_index
            moving_endpoint = moving_edge[moving_endpoint_index]
            origin = from_edge(network, moving_edge, moving_endpoint=moving_endpoint)
            try:
                move = Move(
                    move_type=move_type,
                    origin=origin,
                    moving_edge=moving_edge,
                    target=target_edge,
                )
                check_valid(network, move)
            except (InvalidMoveException, InvalidMoveDefinitionException):
                continue
            yield move


def _all_valid_vplu_moves(network):
    for start_edge in network.edges:
        for end_edge in network.edges:
            try:
                move = Move(
                    move_type=MoveType.VPLU,
                    start_edge=start_edge,
                    end_edge=end_edge,
                    network=network,
                )
                check_valid(network, move)
            except (InvalidMoveException, InvalidMoveDefinitionException) as e:
                continue
            yield move


def _all_valid_vmin_moves(network):
    for removed_edge in network.edges:
        try:
            move = Move(
                move_type=MoveType.VMIN,
                removed_edge=removed_edge,
            )
            check_valid(network, move)
        except (InvalidMoveException, InvalidMoveDefinitionException):
            continue
        yield move


def all_valid_moves(network, move_type=MoveType.ALL):
    """
    Generates all possible valid moves for a given network and move type.

    :param network: a phylogenetic network (phylox.DiNetwork).
    :param move_type: the type of moves to generate (phylox.rearrangement.movetype.MoveType).
    :yield: a move (phylox.rearrangement.move.Move).

    :example:
    >>> from phylox import DiNetwork
    >>> from phylox.rearrangement.move import all_valid_moves
    >>> network = DiNetwork(
    ...     edges=[(0,1),(1,2),(1,3)],
    ... )
    >>> moves = list(all_valid_moves(network, MoveType.TAIL))
    >>> len(moves)
    4
    >>> moves[0].is_type(MoveType.TAIL)
    True
    >>> moves = list(all_valid_moves(network, MoveType.VERT))
    >>> len(moves)
    4
    >>> moves[0].is_type(MoveType.VPLU)
    True
    """
    if move_type in [MoveType.TAIL, MoveType.RSPR, MoveType.ALL]:
        for move in _all_valid_tail_or_head_moves(network, move_type=MoveType.TAIL):
            yield move
    if move_type in [MoveType.HEAD, MoveType.RSPR, MoveType.ALL]:
        for move in _all_valid_tail_or_head_moves(network, move_type=MoveType.HEAD):
            yield move
    if move_type in [MoveType.VPLU, MoveType.VERT, MoveType.ALL]:
        for move in _all_valid_vplu_moves(network):
            yield move
    if move_type in [MoveType.VMIN, MoveType.VERT, MoveType.ALL]:
        for move in _all_valid_vmin_moves(network):
            yield move


class Move(object):
    """
    A move is a rearrangement operation on a phylogenetic network.

    :param move_type: the type of move (phylox.rearrangement.movetype.MoveType).
    :param kwargs: the parameters of the move, depending on the move type.
    """

    def __init__(self, *args, **kwargs):
        try:
            self.move_type = kwargs["move_type"]
        except KeyError:
            raise InvalidMoveDefinitionException("Missing move_type.")

        # None type move
        if self.move_type == MoveType.NONE:
            return

        # TAIL/HEAD move (i.e. RSPR/horizontal)
        elif self.move_type == MoveType.RSPR:
            raise InvalidMoveDefinitionException(
                "rSPR moves must be defined as moves of type tail or head."
            )
        elif self.move_type in [MoveType.TAIL, MoveType.HEAD]:
            try:
                self.moving_edge = kwargs["moving_edge"]
                self.target = kwargs["target"]
            except KeyError:
                raise InvalidMoveDefinitionException(
                    "Missing one of moving_edge or target."
                )
            self.origin = kwargs.get("origin")
            if self.origin is None:
                if "network" not in kwargs:
                    raise InvalidMoveDefinitionException(
                        "Either a network or an origin must be given."
                    )
                moving_endpoint_index = 0 if self.move_type == MoveType.TAIL else 1
                moving_endpoint = self.moving_edge[moving_endpoint_index]
                try:
                    self.origin = from_edge(
                        kwargs["network"],
                        self.moving_edge,
                        moving_endpoint=moving_endpoint,
                    )
                except NetworkXError:
                    raise InvalidMoveDefinitionException(
                        "Origin edge could not be computed."
                    )
            if self.move_type == MoveType.TAIL:
                self.moving_node = self.moving_edge[0]
            else:
                self.moving_node = self.moving_edge[1]

            if self.moving_edge == self.target:
                raise InvalidMoveDefinitionException(
                    "Moving edge must not be the target edge."
                )

            return
        # VERT move (i.e. SPR/vertical)
        elif self.move_type == MoveType.VERT:
            raise InvalidMoveDefinitionException(
                "vertical moves must be defined as moves of type VPLU or VMIN."
            )
        # VPLU/VMIN move
        elif self.move_type == MoveType.VPLU:
            try:
                self.start_edge = kwargs["start_edge"]
                self.end_edge = kwargs["end_edge"]
                self.start_node = kwargs.get("start_node", None)
                self.end_node = kwargs.get("end_node", None)
                network = kwargs.get("network", None)
                if (
                    self.start_node is None or self.end_node is None
                ) and network is None:
                    raise InvalidMoveDefinitionException(
                        "Either a start_node and end_node, or a network must be given."
                    )
                if self.start_node is None:
                    self.start_node = find_unused_node(network)
                if self.end_node is None:
                    self.end_node = find_unused_node(network, exclude=[self.start_node])

                if self.start_edge == self.end_edge:
                    raise InvalidMoveDefinitionException(
                        "Start edge must not be the end edge."
                    )
            except KeyError:
                raise InvalidMoveDefinitionException(
                    "Missing one of start_edge, end_edge."
                )
        elif self.move_type == MoveType.VMIN:
            try:
                self.removed_edge = kwargs["removed_edge"]
            except KeyError:
                raise InvalidMoveDefinitionException(
                    "Missing removed_edge in definition."
                )
        else:
            raise InvalidMoveDefinitionException("Invalid move type.")

    def is_type(self, move_type):
        """
        Checks if the move is of a given type.

        :param move_type: the move type to check (phylox.rearrangement.movetype.MoveType).
        :return: True if the move is of the given type, False otherwise.

        :example:
        >>> from phylox.rearrangement.move import Move
        >>> move = Move(
        ...     move_type=MoveType.TAIL,
        ...     origin=(2,5),
        ...     moving_edge=(1,3),
        ...     target=(2,4),
        ... )
        >>> move.is_type(MoveType.TAIL)
        True
        >>> move.is_type(MoveType.RSPR)
        True
        >>> move.is_type(MoveType.HEAD)
        False
        >>> move.is_type(MoveType.VERT)
        False
        """

        if move_type == MoveType.ALL:
            return True
        if (
            self.move_type == MoveType.NONE
            or (
                move_type == MoveType.RSPR
                and self.move_type in [MoveType.TAIL, MoveType.HEAD]
            )
            or (
                move_type == MoveType.VERT
                and self.move_type in [MoveType.VPLU, MoveType.VMIN]
            )
        ):
            return True
        return move_type == self.move_type

    def rename_nodes(self, isomorphism):
        """
        Renames the nodes in the move using an isomorphism mapping between two networks.

        :param isomorphism: a dictionary, containing a bijective mapping from nodes of the networks to another set.
        :return: a new move with the nodes renamed.
        """
        if self.move_type == MoveType.NONE:
            return self
        if self.move_type in [MoveType.TAIL, MoveType.HEAD]:
            return Move(
                move_type=self.move_type,
                origin=(isomorphism[self.origin[0]], isomorphism[self.origin[1]]),
                moving_edge=(
                    isomorphism[self.moving_edge[0]],
                    isomorphism[self.moving_edge[1]],
                ),
                target=(isomorphism[self.target[0]], isomorphism[self.target[1]]),
            )
        elif self.move_type == MoveType.VPLU:
            return Move(
                move_type=self.move_type,
                start_edge=(
                    isomorphism[self.start_edge[0]],
                    isomorphism[self.start_edge[1]],
                ),
                end_edge=(isomorphism[self.end_edge[0]], isomorphism[self.end_edge[1]]),
                start_node=isomorphism[self.start_node],
                end_node=isomorphism[self.end_node],
            )
        elif self.move_type == MoveType.VMIN:
            return Move(
                move_type=self.move_type,
                removed_edge=(
                    isomorphism[self.removed_edge[0]],
                    isomorphism[self.removed_edge[1]],
                ),
            )

    def invert(self, network=None):
        """
        Inverts the move.

        :param network: a phylogenetic network (phylox.DiNetwork) in which to invert the move.
            (needed for VMIN moves)
        :return: a new move that is the inverse of the current move.
        """
        if self.move_type == MoveType.NONE:
            return self
        if self.move_type in [MoveType.TAIL, MoveType.HEAD]:
            return Move(
                move_type=self.move_type,
                origin=self.target,
                moving_edge=self.moving_edge,
                target=self.origin,
            )
        elif self.move_type == MoveType.VPLU:
            return Move(
                move_type=self.VMIN,
                removed_edge=(self.start_node, self.end_node),
            )
        elif self.move_type == MoveType.VMIN:
            if network is None:
                raise InvalidMoveException(
                    "Cannot invert a VMIN move without a network."
                )
            start_edge = from_edge(network, self.removed_edge, self.removed_edge[0])
            end_edge = from_edge(network, self.removed_edge, self.removed_edge[1])
            return Move(
                move_type=self.VPLU,
                start_edge=start_edge,
                end_edge=end_edge,
                start_node=self.removed_edge[0],
                end_node=self.removed_edge[1],
            )

    @classmethod
    @np_random_state("seed")
    def random_move(
        cls,
        network,
        available_tree_nodes=None,
        available_reticulations=None,
        move_type_probabilities={
            MoveType.TAIL: 0.4,
            MoveType.HEAD: 0.4,
            MoveType.VPLU: 0.1,
            MoveType.VMIN: 0.1,
        },
        seed=None,
    ):
        """
        Generates a random move for the given network.
        The move may not be valid.

        :param network: a phylogenetic network (phylox.DiNetwork).
        :param available_tree_nodes: a list of available tree nodes to use for the move.
        :param available_reticulations: a list of available reticulations to use for the move.
        :param move_type_probabilities: a dictionary of move type probabilities.
        :return: a random move (phylox.rearrangement.move.Move).

        :example:
        >>> from phylox import DiNetwork
        >>> from phylox.rearrangement.move import Move
        >>> network = DiNetwork(
        ...     edges=[(0,1),(1,2),(1,3),(2,3),(2,4),(3,5)],
        ... )
        >>> move = Move.random_move(
        ...     network,
        ...     move_type_probabilities={
        ...         MoveType.TAIL: 0.5,
        ...         MoveType.HEAD: 0.5,
        ...     },
        ... )
        >>> move.is_type(MoveType.RSPR)
        True
        """

        available_tree_nodes = available_tree_nodes or []
        available_reticulations = available_reticulations or []
        edges = list(network.edges())
        num_edges = len(edges)
        move_type_probabilities_keys = list(move_type_probabilities.keys())
        movetype_index = seed.choice(
            len(move_type_probabilities_keys), p=tuple(move_type_probabilities.values())
        )
        movetype = move_type_probabilities_keys[movetype_index]
        if movetype in [MoveType.TAIL, MoveType.HEAD]:
            moving_edge_index = seed.choice(num_edges)
            target_index = seed.choice(num_edges - 1)
            if target_index >= moving_edge_index:
                target_index += 1
                target_index %= num_edges
            moving_edge = edges[moving_edge_index]
            target = edges[target_index]
            moving_endpoint_index = 0 if movetype == MoveType.TAIL else 1
            moving_endpoint = moving_edge[moving_endpoint_index]
            origin = from_edge(network, moving_edge, moving_endpoint=moving_endpoint)
            return Move(
                move_type=movetype,
                origin=origin,
                moving_edge=moving_edge,
                target=target,
            )
        elif movetype == MoveType.VPLU:
            available_reticulations = list(available_reticulations) or [
                find_unused_node(network)
            ]
            available_tree_nodes = list(available_tree_nodes) or [
                find_unused_node(network, exclude=available_reticulations)
            ]
            start_edge = edges[seed.choice(num_edges)]
            end_edge = edges[seed.choice(num_edges)]
            start_node = seed.choice(available_tree_nodes)
            end_node = seed.choice(available_reticulations)
            return Move(
                move_type=movetype,
                start_edge=start_edge,
                end_edge=end_edge,
                start_node=start_node,
                end_node=end_node,
            )
        elif movetype == MoveType.VMIN:
            removed_edge = edges[seed.choice(num_edges)]
            return Move(move_type=movetype, removed_edge=removed_edge)
